import numpy as np
import math
from sys import getsizeof

from pyplan_core.cubepy.axes import make_axes
from pyplan_core.cubepy.axis import Axis
from pyplan_core.cubepy.exceptions import AxisAlignError
from pyplan_core.cubepy.index import Index
from pyplan_core.cubepy.utils import make_axis_collection, is_axis, is_indexed, align_arrays, broadcast_array, unique_axes_from_cubes
from pyplan_core.cubepy.exceptions import InvalidAxisLengthError

class Cube(object):
    """Wrapper around numpy.ndarray with named and labelled axes. The API aims to be as similar to ndarray API as
    possible. Moreover it allows automatic axis matching and alignment in operations among cubes.
    """

    # when numpy array is the first argument in operation and Cube is the second,
    # then __array_priority__ will force Cube to handle the operation rather than numpy array
    __array_priority__ = 10

    def __init__(self, axes, values, dtype=None, fillValues=True):
        if not isinstance(values, np.ndarray):
            # masked arrays will not be affected
            values = np.asarray(values, dtype)

        axes = make_axes(axes)

        def tryRecreateValues(oldValues,axes):
            factor=1
            for x in axes:
                factor = factor * len(x)
            axisShape = [len(x) for x in axes]
            if factor-oldValues.size > 0:
                if oldValues.dtype.kind in {'U', 'S'}:
                    return np.append(oldValues.reshape(oldValues.size),['']* (factor-oldValues.size)).reshape(axisShape)
                elif oldValues.dtype.kind in {'b'}:
                    return np.append(oldValues.reshape(oldValues.size),[False]* (factor-oldValues.size)).reshape(axisShape)
                else:
                    return np.append(oldValues.reshape(oldValues.size),np.zeros(factor-oldValues.size)).reshape(axisShape)
            else:
                return oldValues.reshape(oldValues.size)[:factor].reshape(axisShape)

        # if axes dimensions and value dimension do not match
        if values.ndim != len(axes):
            if fillValues:
                values = tryRecreateValues(values,axes)
            else:
                raise ValueError("invalid number of axes")

        lenIsOk=True
        for n, axis in zip(values.shape, axes):
            if n != len(axis):
                lenIsOk=False
        if not lenIsOk:
            if fillValues:
                values = tryRecreateValues(values,axes)
            else:
                raise InvalidAxisLengthError("invalid length of axis '{}'".format(axis.name))


        if values.dtype.name=="int32":
            self._values = values.astype(np.int64)
        else:
            self._values = values
        self._axes = axes



    def __getitem__(self, items, fillWithNan=True):
        """Similar rules apply as with indexing and slicing numpy ndarray.
        Notes:
        1) np.newaxis is not supported.
        2) axes indexed by integer are collapsed
        :param item:
        :return: new Cube instance
        """
        if isinstance(items,self.__class__):
            if items.ndim == 1:
                values = items.axes[0].values[items.values]
                res = self.filter(items.axes[0].name,values)
                if res.ndim==1 and len(res.values)==1:
                    return res.values.tolist()[0]
                else:
                    return res.squeeze(items.axes[0].name)
            else:
                
                ohterDims = []
                for dim in items.dims:
                    if not dim in self.dims: 
                        ohterDims.append(dim)
                        
                intersect = []
                for dim in items.dims:
                    if  dim in self.dims: 
                        intersect.append(dim)        

                if items.ndim == 2 and len(ohterDims)==1 and len(intersect)==1:
                    
                    old_index = items.axis(intersect[0])
                    new_index = items.axis(ohterDims[0])
                    cube = self
                    old_index_id = intersect[0]
                    new_index_id = ohterDims[0]

                    #condition  matrix 
                    _condition = items #old_index == new_index  
                    _filtered = _condition.any(keep=old_index_id)
                    _filterArray  = _filtered[_filtered==True]

                    #fix only one filter
                    if isinstance(_filterArray,bool):
                        print(type(_filterArray))
                        print( old_index.values[_filtered.values] )
                        _boolIndex = Index(old_index_id, old_index.values[_filtered.values] )
                        _filterArray = self.__class__([_boolIndex],[True])

                    #filtered cube
                    _preCube = cube.filter(_filterArray)
                    
                    #values not in old index
                    _mask = np.in1d(new_index.values,old_index.values, invert=True)
                    _notinLen = np.sum(_mask)

                    _axis_index = self.axis_index(old_index_id)
                    resultCube = None
                    #have other values than old index
                    if _notinLen>0:

                        _tmpIndexValues = np.concatenate([_preCube.axis(old_index_id).values, new_index.values[_mask]])
                        _tmpIndex = Index("_tmp_index_", _tmpIndexValues )
                        _arrDimLen = []
                        _arrDim = []
                        for dim in cube.axes:
                            if dim.name == old_index_id:
                                _arrDimLen.append(_notinLen)
                                _arrDim.append(_tmpIndex)
                            else:
                                _arrDimLen.append( len(dim))
                                _arrDim.append(dim)
                    

                        _extraValues = None
                        if fillWithNan:
                            _extraValues = np.empty(_arrDimLen, dtype=object)
                            _extraValues[:] = np.nan
                        else:
                            _extraValues = np.empty(_arrDimLen, dtype=self.values.dtype)
                            #try deduce data type for empty data for object data types
                            if self.values.dtype.kind.lower()=="o":
                                if len(self.values)>0:
                                    _extraValues[:] = self.values.flatten()[0] * False
                            elif kindToString(self.values.dtype.kind)=="numeric":
                                _extraValues[:]=0
                            elif kindToString(self.values.dtype.kind)=="string":
                                _extraValues[:]=""


                        _newValues =  _preCube.values
                        _finalValues = np.concatenate( [ _newValues,_extraValues], axis=_axis_index)
                        
                        resultCube = self.__class__(_arrDim,_finalValues, dtype=self.values.dtype)
                    else:
                        resultCube = _preCube.rename_axis(old_index_id,"_tmp_index_")


                    # set values on correct order an replace axis
                    _sourceValues = resultCube.axis("_tmp_index_").values
                    _targetValues = new_index.values

                    #check if is necesary reorder values
                    if not np.array_equal(_sourceValues, _targetValues):
                        _sortedBy = [np.where(_sourceValues == x)[0][0]  for x in _targetValues] 
                        resultCube._values = resultCube.values.take(_sortedBy, axis=_axis_index)
                    res = resultCube.replace_axis("_tmp_index_",new_index)
                else:
                    def _subscriptWithoutSqueeze(selfCube,condition):
                        _values = condition.axes[0].values[condition.values]
                        _res = selfCube.filter(condition.axes[0].name,_values)
                        return _res

                    filtered = items.reduce(safesum, axis=ohterDims)
                    res=self
                    if filtered.ndim==1:
                        res = _subscriptWithoutSqueeze(self,filtered>0)
                    else:
                        for axis in filtered.axes:
                            res = _subscriptWithoutSqueeze(res,filtered.sum(axis)>0)

                        res = (res * filtered).squeeze()

                return res
            #else:
             #   raise ValueError("Subscript is only supported with 1 or 2 dimensions")
        else:
            if not isinstance(items, tuple):
                items = (items,)
            new_axes = []

            # append the axes given by items
            for item, axis in zip(items, self._axes):
                if not isinstance(item, int):  # indexing by int collapses a dimension
                    new_axes.append(axis[item])

            # append the rest of axes
            for i in range(len(items), len(self._axes)):
                new_axes.append(self._axes[i])

            return self.__class__(new_axes, self._values[items])

    def __bool__(self):
        """Return the truth value of the cube.
        If the cube is empty, returns False.
        If the cube is scalar, returns the truth value of the only element.
        If the cube has more than one element, ValueError is raised.
        Note: The function returns the truth value of the underlying numpy ndarray.
        """
        return bool(self._values)

    def __repr__(self):
        """Returns a textual representation of the object.
        :return: str
        """
        return "Cube({}, {})".format(repr(tuple(self.axes)),self._values)

    def __sizeof__(self):
        res = self._values.nbytes
        for axis in self._axes:
            res += getsizeof(axis)
        return res

    def __len__(self):
        """Returns the number of elements in the underlying numpy.ndarray.
        :return: int
        """
        return self.size

    @property
    def shape(self):
        """Returns the lengths of dimensions of the underlying numpy.ndarray.
        :return: tuple of ints
        """
        return self._values.shape

    @property
    def size(self):
        """Returns the number of elements in the underlying numpy.ndarray.
        :return: int
        """
        return self._values.size

    @property
    def ndim(self):
        """Returns the number of array dimensions.
        :return: int
        """
        return self._values.ndim

    @property
    def values(self):
        return self._values  # TODO: .view()?

    @property
    def axes(self):
        """Returns a tuple of Axis objects."""
        return self._axes.axes

    @property
    def dims(self):
        """Returns a tuple of axis names."""
        return self._axes.dims

    def axis(self, axis):
        """Returns axis by the name or by the index.
        Index can be a negative number, in that case, the axes are counted backwards from the last one.
        :param axis: axis name (str), index (int) or instance
        :return: Axis instance
        :raise LookupError: if the axis does not exist, TypeError if wrong argument type is passed
        """
        return self._axes[self.axis_index(axis)]

    def axis_index(self, axis):
        """Returns the index of the axis specified by its name or axis instance.
        :param axis: axis name (str), index (int) or instance
        :return: int
        :raise LookupError: if the axis does not exist, TypeError if wrong argument type is passed
        """
        return self._axes.index(axis)

    def has_axis(self, axis):
        """Returns True/False indicating whether the axis exists in the Cube.
        :param axis: axis name (str), index (int) or instance
        :return: bool
        :raise TypeError: if wrong argument type is passed
        """
        return self._axes.contains(axis)

    def apply(self, func, *args):
        """Applies a function to each element individually and return the new cube with the same dimensions.
        :param func: function to be applied to values
        :param args: additional optional arguments of func
        :return: new Cube instance

        Examples:
        cube.apply(np.sin)
        cube.apply(np.percentile, 10)  # i.e. 1st decile
        cube.apply(lambda x: x ^ 2 if x > 0 else 0)  # quadratic function for positive values, otherwise zero
        """

        # now with vectorize for ever
        func = np.vectorize(func)
        values = func(self._values, *args)
        return self.__class__(self._axes,values)


    def transpose(self, front=[], back=[]):
        """A generalized analogy to numpy.transpose.
        :param front: axes which will be in the front of other axes
        :param back: axes which will be at the back of other axes
        :return: new Cube instance with transposed axes

        The arguments 'front' and 'back' are expected in the form of an axis identifier or a collection
        of axis identifiers. Axis identifier is a name (str), index (int) or Axis instance.
        """
        indices = self._axes.transposed_indices(front, back)
        new_axes = tuple(self._axes.axis_by_index(index) for index in indices)
        new_values = self._values.transpose(indices)
        return self.__class__(new_axes,new_values)

    def squeeze(self, axis=None):
        """Removes all the axes with the size equal to 1 from the cube.
        Analogy to numpy ndarray.squeeze().
        :param axis: optional axis name (str), index (int) or instance to perform squeeze only in this
        :return: new Cube instance
        """
        new_axes = None
        if axis is None:
            new_axes = tuple(a for a in self.axes if len(a) != 1)
        else:
            axisInstance = self.axis(axis)
            new_axes = tuple(a for a in self.axes if len(a) != 1 or not a is axisInstance)

        new_values = self._values.squeeze()
        return self.__class__(new_axes,new_values)
        
    # ****************************
    # *** Arithmetic operators ***
    # ****************************

    # unary +
    def __pos__(self):
        return self

    # unary -
    def __neg__(self):
        return self.__class__(self._axes,-self._values)

    # A + B
    def __add__(self, other):
        _fn = self.selectOptimalFunction(self, other, np.add)
        return apply_op(self, other, _fn)

    def __radd__(self, other):
        _fn = self.selectOptimalFunction(other,self, np.add)
        return apply_op(other, self, _fn)

    # A * B
    def __mul__(self, other):
        _fn = self.selectOptimalFunction(self, other, np.multiply)
        return apply_op(self, other, _fn)

    def __rmul__(self, other):
        _fn = self.selectOptimalFunction(other, self, np.multiply)
        return apply_op(other, self, _fn)

    # A - B
    def __sub__(self, other):
        return apply_op(self, other, np.subtract)

    def __rsub__(self, other):
        return apply_op(other, self, np.subtract)

    # A / B - division for Python 2
    # if both operands are int then result is int, otherwise it is float
    def __div__(self, other):
        return apply_op(self, other, np.divide)

    def __rdiv__(self, other):
        return apply_op(other, self, np.divide)

    # A / B - division for Python 3
    # result is always float even if both operands are int
    def __truediv__(self, other):
        return apply_op(self, other, np.true_divide)

    def __rtruediv__(self, other):
        return apply_op(other, self, np.true_divide)

    # A // B - divide and floor down
    # if both operands are int then result is int, otherwise it is float (for both Python 2 and Python 3)
    def __floordiv__(self, other):
        return apply_op(self, other, np.floor_divide)

    def __rfloordiv__(self, other):
        return apply_op(other, self, np.floor_divide)

    # A ** B
    def __pow__(self, other):
        return apply_op(self, other, np.power)

    def __rpow__(self, other):
        return apply_op(other, self, np.power)

    # A % B (modulo)
    def __mod__(self, other):
        return apply_op(self, other, np.mod)

    def __rmod__(self, other):
        return apply_op(other, self, np.mod)

    # *************************
    # *** Bitwise operators ***
    # *************************

    def __invert__(self):
        """Returns bit-wise inversion, or bit-wise NOT, element-wise."""
        if kindToString(self._values.dtype.kind)=="numeric":
            return self.__class__(self._axes,np.invert(self._values!=0)*1)
        else:
            return self.__class__(self._axes,np.invert(self._values))

    # A & B
    def __and__(self, other):
        return (apply_op(self, other, np.bitwise_and)==1)

    def __rand__(self, other):
        return (apply_op(other, self, np.bitwise_and)==1)

    # A | B
    def __or__(self, other):
        return (apply_op(self, other, np.bitwise_or)==1)

    def __ror__(self, other):
        return (apply_op(other, self, np.bitwise_or)==1)

    # A ^ B
    def __xor__(self, other):
        return apply_op(self, other, np.bitwise_xor)

    def __rxor__(self, other):
        return apply_op(other, self, np.bitwise_xor)

    # A >> B
    def __lshift__(self, other):
        return apply_op(self, other, np.left_shift)

    def __rlshift__(self, other):
        return apply_op(other, self, np.left_shift)

    # A << B
    def __rshift__(self, other):
        return apply_op(self, other, np.right_shift)

    def __rrshift__(self, other):
        return apply_op(other, self, np.right_shift)

    # ****************************
    # *** Comparison operators ***
    # ****************************

    # A == B
    def __eq__(self, other):
        return apply_op(self, other, np.equal)

    # A != B
    def __ne__(self, other):
        return apply_op(self, other, np.not_equal)

    # A < B
    def __lt__(self, other):
        return apply_op(self, other, np.less)

    # A <= B
    def __le__(self, other):
        return apply_op(self, other, np.less_equal)

    # A > B
    def __gt__(self, other):
        return apply_op(self, other, np.greater)

    # A >= B
    def __ge__(self, other):
        return apply_op(self, other, np.greater_equal)

    # ***************************************
    # *** Built-in mathematical functions ***
    # ***************************************

    def __abs__(self):
        """Implements behaviour for the built in abs() function.
        :return: new Cube instance
        """
        return self.apply(abs)

    def __round__(self, decimals):
        """Implements behaviour for the built in round() function.
        :param decimals: the number of decimal places to round to
        :return: new Cube instance
        """
        return self.apply(round, decimals)

    def __floor__(self):
        """Implements behaviour for math.floor(), i.e., rounding down to the nearest integer.
        :return: new Cube instance
        """
        return self.apply(math.floor)

    def __ceil__(self):
        """Implements behaviour for math.ceil(), i.e., rounding up to the nearest integer.
        :return: new Cube instance
        """
        return self.apply(math.ceil)

    def __trunc__(self):
        """Implements behavior for math.trunc(), i.e., truncating to an integral.
        :return: new Cube instance
        """
        return self.apply(math.trunc)


    def argsort(self,axis):
        """Return new cube with the position of values sorted by axis
        """
        _, axis_index = self._axis_and_index(axis)
        new_values = np.argsort(self._values, axis=axis_index)
        return self.__class__(self.axes,new_values)

    # ************************************
    # *** Numpy mathematical functions ***
    # ************************************

    # The following applies to most of the functions listed below:
    # 1) The function is performed element-wise, the axes are left untouched and a new instance of cube is returned.
    # 2) Function func can be applied on cube c by calling np.func(c) or c.func().

    def sin(self):
        """Sine."""
        return self.apply(np.sin)

    def cos(self):
        """Cosine."""
        return self.apply(np.cos)

    def tan(self):
        """Tangent."""
        return self.apply(np.tan)

    def exp(self):
        """Exponential."""
        return self.apply(np.exp)

    def log(self):
        """Logarithm with natural base."""
        return self.apply(np.log)

    def log2(self):
        """Logarithm with base of 2."""
        return self.apply(np.log2)

    def log10(self):
        """Logarithm with base of 10."""
        return self.apply(np.log10)

    def isnan(self):
        """Returns True for all NaN values and False for all non-NaN values."""
        return self.apply(np.isnan)

    def clearNan(self, defaultValue=0):
        """ Apply lambda x: defaultValue if np.isnan(x) else x """
        _kind = kindToString(self.values.dtype.kind)
        if _kind == "numeric":
            if defaultValue==0:
                if self.values.dtype.kind == "f":
                    defaultValue=0.
        return self.apply(lambda x: defaultValue if np.isnan(x) else x)

    def clearInf(self, defaultValue=0):
        """ Apply lambda x: defaultValue if np.isfinite(x) else x """
        _kind = kindToString(self.values.dtype.kind)
        if _kind == "numeric":
            if defaultValue==0:
                if self.values.dtype.kind == "f":
                    defaultValue=0.
        return self.apply(lambda x: defaultValue if np.isinf(x) else x)
    
    def clearAll(self, defaultValue=0):
        """ Apply lambda x: defaultValue if np.isfinite(x) or np.isnan(x) else x """
        _kind = kindToString(self.values.dtype.kind)
        if _kind == "numeric":
            if defaultValue==0:
                if self.values.dtype.kind == "f":
                    defaultValue=0.

        return self.apply(lambda x: defaultValue if np.isinf(x) or np.isnan(x) else x)


    # *****************************
    # *** Aggregation functions ***
    # *****************************

    def sum(self, axis=None, keep=None, group=None, sort_grp=True):
        """Sum of array elements over a given axis.

        :param axis: Axis or axes along which a sum is performed. The default (axis = None) is perform a sum
        over all the dimensions of the input array. axis may be negative, in which case it counts from the last
        to the first axis. If this is a tuple of ints, a sum is performed on multiple axes, instead of a single
        axis or all the axes as before.
        :return: new Cube instance or a scalar value
        """
        try:
            return self.reduce(np.sum, axis, keep, group, sort_grp)
        except (ValueError, TypeError):
            return self.reduce(safesum, axis, keep, group, sort_grp)

        

    def mean(self, axis=None, keep=None, group=None, sort_grp=True):
        """Returns the arithmetic mean."""
        return self.reduce(np.mean, axis, keep, group, sort_grp)

    def nanmean(self, axis=None, keep=None, group=None, sort_grp=True):
        """Returns the arithmetic mean with exclusion of NaN values."""
        return self.reduce(np.nanmean, axis, keep, group, sort_grp)

    def median(self, axis=None, keep=None, group=None, sort_grp=True):
        """Returns the median (i.e. the middle value)."""
        return self.reduce(np.median, axis, keep, group, sort_grp)

    def min(self, axis=None, keep=None, group=None, sort_grp=True):
        """Returns the minimum values."""
        return self.reduce(np.min, axis, keep, group, sort_grp)

    def max(self, axis=None, keep=None, group=None, sort_grp=True):
        """Returns the maximum value."""
        return self.reduce(np.max, axis, keep, group, sort_grp)

    def all(self, axis=None, keep=None, group=None, sort_grp=True):
        """Tests whether all elements evaluate to True."""
        return self.reduce(np.all, axis, keep, group, sort_grp)

    def any(self, axis=None, keep=None, group=None, sort_grp=True):
        """Tests whether any element evaluates to True."""
        return self.reduce(np.any, axis, keep, group, sort_grp)

    def prod(self, axis=None, keep=None, group=None, sort_grp=True):
        """Returns the product (i.e. multiplication) of the values."""
        return self.reduce(np.prod, axis, keep, group, sort_grp)

    def count_nonzero(self, axis=None, keep=None, group=None, sort_grp=True):
        return self.reduce(np.count_nonzero, axis, keep, group, sort_grp)

    def std(self, axis=None, keep=None, group=None, sort_grp=True):
        """Returns the standard deviation."""
        return self.reduce(np.std, axis, keep, group, sort_grp)

    def nanstd(self, axis=None, keep=None, group=None, sort_grp=True):
        """Returns the standard deviation with exclusion of NaN values."""
        return self.reduce(np.nanstd, axis, keep, group, sort_grp)

    def var(self, axis=None, keep=None, group=None, sort_grp=True):
        """Returns the variance."""
        return self.reduce(np.var, axis, keep, group, sort_grp)

    def nanvar(self, axis=None, keep=None, group=None, sort_grp=True):
        """Returns the variance with exclusion of NaN values."""
        return self.reduce(np.nanvar, axis, keep, group, sort_grp)

    def reduce(self, func, axis=None, keep=None, group=None, sort_grp=True):
        """Aggregation of values in the cube along one or more axes. This function works
        in three different modes:
         1) the axes to be eliminated are specified
         2) the axes to be kept are specified, while the other axes are eliminated
         3) values are grouped along a specified axis

        :param func: the function which is used to aggregate the values
            It must take two values
        :param axis: axis or axes to be eliminated by the aggregation
        :param keep: axis or axes which are kept after the aggregation
        :param group: axis for which values are the results grouped
        :param sort_grp: True to sort the grouped values, False to keep the order of the first occurrences
            This is applicable only when 'group' is defined

        No more than one of 'axis', 'keep' and 'group' arguments can be defined (i.e. non-None), otherwise
        ValueError is raised. If none of these is defined, then the Cube is aggregated to a single scalar value.

        Example:
        # returns sum of all months, i.e. month axis is eliminated; other axes are kept
        cube.aggregate(np.sum, "month")

        # returns mean for each month, i.e. month axis is kept; other axes are eliminated
        cube.aggregate(np.mean, keep="month")
        """

        aggr_params = int(axis is not None) + int(keep is not None) + int(group is not None)
        if aggr_params == 0:
            # total aggregation from cube to scalar
            return func(self._values)
        elif aggr_params > 1:
            raise ValueError("no more than one of 'axis', 'keep' or 'group' arguments can be defined")

        if axis is not None or keep is not None:
            axis = make_axis_collection(axis)
            keep = make_axis_collection(keep)

            if axis is not None:
                axis_indices_to_remove = tuple(self._axes.index(a) for a in axis)
                new_axes = list(a for i, a in enumerate(self._axes) if i not in axis_indices_to_remove)
            else:
                axis_index_set = set(self._axes.index(a) for a in keep)
                new_axes = list(a for i, a in enumerate(self._axes) if i in axis_index_set)
                axis_indices_to_remove = tuple(set(range(self.ndim)) - axis_index_set)
            return self._aggregate(func, new_axes, axis_indices_to_remove)

        elif group is not None:
            return self._group(group, func, sort_grp)

    def diff(self, axis, n=1, axis_shift=None):
        """Calculate the n-th order discrete difference along given axis.
        The first order difference is given by out[n] = a[n+1] - a[n] along the given axis,
        higher order differences are calculated by using diff recursively.
        :param axis: axis name (str), index (int) or instance
        :param n: difference order
        :param axis_shift: by how many values is the new axis to be shifted; by default is equal to n
        :return: new Cube instance
        """
        if axis_shift is None:
            axis_shift = n
        old_axis, axis_index = self._axis_and_index(axis)
        new_values = np.diff(self.values, n=n, axis=axis_index)
        new_length = len(old_axis) - n
        new_axis = old_axis[axis_shift: (axis_shift + new_length)]
        new_axes = self._axes.replace(axis_index, new_axis)
        return self.__class__(new_axes,new_values)


    def cumProd(self, axis):
        """Calculate the cumprod over specified axis
        :param axis: axis name (str), index (int) or instance
        :return: new Cube instance
        """
        _ax, axis_index = self._axis_and_index(axis)
        new_values = np.cumprod(self.values, axis=axis_index)
        return self.__class__(self._axes,new_values)        


    def growth(self, axis, n=1, axis_shift=None):
        """Returns relative growth rate in the direction of specified axis.
        :param axis: axis name (str), index (int) or instance
        :param n: growth order
        :param axis_shift: by how many values is the new axis to be shifted; by default is equal to n
        :return: new Cube instance
        Example:
        If you have quarterly values by you want to calculate year-to-year index, use n=4.
        """
        if axis_shift is None:
            axis_shift = n
        old_axis, axis_index = self._axis_and_index(axis)
        new_length = len(old_axis) - n
        forth_values = self.last(axis_index, new_length).values
        back_values = self.first(axis_index, new_length).values
        new_values = forth_values / back_values
        new_axis = old_axis[axis_shift: (axis_shift + new_length)]
        new_axes = self._axes.replace(axis_index, new_axis)
        return self.__class__(new_axes,new_values)

    def masked(self, func):
        """Returns a cube with masked values.
        :param func: function which is applied to each
        :return: new Cube instance with masked values
        Example: to calculate mean of non-Nan values
        r = cube.masked(np.isnan).mean()
        """
        mask = self.apply(func)._values
        masked_values = np.ma.masked_array(self._values, mask)
        return self.__class__(self._axes,masked_values)


    def shift(self, axis, n=1, cval=0):
        """Returns a cube with the axis shifted."""

        n=n*-1
        axis, axis_index = self._axis_and_index(axis)
        newValues=np.roll(self.values,n ,axis=axis_index)
        newCube = self.__class__(self._axes,newValues)
    
        # set default values    
        indexToFillCVal = None
        if n>0:
            indexToFillCVal=axis[:n]
        else:
            indexToFillCVal=axis[n:]
        slices = [slice(None)] * self.ndim
        wh = None
        if self.has_axis(indexToFillCVal.name):
            wh = np.where(np.isin(axis.values, indexToFillCVal.values ))
            if (len(wh)>0):
                slices[axis_index] = wh[0]
            newCube.values[tuple(slices)] = cval
        #end set default values

        return newCube
    

    # **************************
    # *** Axes manipulations ***
    # **************************

    def replace_axis(self, old_axis, new_axis):
        """Replaces an existing axis with a new axis of the same length and returns the new Cube instance.
        The new axis may have different name but it must be unique among the other axes.
        :param old_axis: axis name (str), index (int) or instance
        :param new_axis: axis instance to replace the old axis
        :return: new Cube instance
        :raise InvalidAxisLengthError: if the new axis has wrong length
        :raise NonUniqueDimNamesError: if the new axis has a name which already exists in the cube
        """
        new_axes = self._axes.replace(old_axis, new_axis)
        return self.__class__(new_axes,self._values)

    def swap_axes(self, axis1, axis2):
        """Swaps two axes.
        :param axis1: axis name (str), index (int) or instance
        :param axis2: axis name (str), index (int) or instance
        :return: new Cube instance with swapped axes
        :raise LookupError: if axis1 or axis2 is not found
        If axis1 is the same as axis2, the original Cube instance is returned.
        """
        index1 = self._axes.index(axis1)
        index2 = self._axes.index(axis2)
        if index1 == index2:
            return self
        new_axes = self._axes.swap(index1, index2)
        new_values = self._values.swapaxes(index1, index2)
        return self.__class__(new_axes,new_values)

    def insert_axis(self, axis, index=0):
        """Adds a new axis and repeats the values to fill the new cube.
        :param axis: the new axis to be inserted
        :param index: the index of the new axis after it is inserted
        :return: new Cube instance with inserted axis
        :raise: TODO
        """
        new_axes = self._axes.insert(axis, index)
        new_values = np.expand_dims(self._values, index)
        new_values = np.repeat(new_values, repeats=len(axis), axis=index)
        return self.__class__(new_axes,new_values)

    def align(self, align_to):
        """Make all matching axes aligned to the given axes.
        :param align_to: Axis instance, Cube instance, collection of Axis or Cube instances
        :return: new Cube instance
        If called with a Cube instance, it is ensured that after this function
        the both cubes can be used in an operation. Moreover there is no need for
        alignment in the operation because the matching axes are identical.
        """
        if is_axis(align_to):
            if self.has_axis(align_to.name):
                return self._align_axis(align_to)
            else:
                return self
        elif is_cube(align_to):
            axes = align_to.axes
        else:
            axes = align_to

        result = self
        for axis in axes:
            result = result.align(axis)
        return result

    def extend(self, axis, fill):
        # TODO...
        pass

    def rename_axis(self, old_axis, new_name):
        """Returns a cube with a renamed axis.
        :param old_axis: axis name (str), index (int) or instance
        :param new_name: the name of the new axis (str)
        :return: new Cube instance
        :raise LookupError: if the old axis does not exist, ValueError is the name is duplicate
        """
        new_axes = self._axes.rename(old_axis, new_name)
        return self.__class__(new_axes,self._values)

    def combine_axes(self, axis_names, new_axis_name, format):
        count = len(axis_names)
        axes = list()
        array_list = list()
        size = 1
        axis_indices = list()
        unique_axis_indices = set()
        for axis_name in axis_names:
            axis, axis_index = self._axis_and_index(axis_name)
            unique_axis_indices.add(axis_index)
            axis_indices.append(axis_index)
            axes.append(axis)
            array_list.append(axis.values)
            size *= len(axis)

        if len(unique_axis_indices) != len(axis_names):
            raise ValueError("axis names are not unique")

        other_indices = list()
        new_axes = list()
        for i, a in enumerate(self._axes):
            if i not in axis_indices:
                if a.name == new_axis_name:
                    raise ValueError("axis name '{}' is not unique".format(new_axis_name))
                other_indices.append(i)
                new_axes.append(a)

        axis_indices.extend(other_indices)
        axis_sizes = [len(self.axis(i)) for i in other_indices]
        axis_sizes.insert(0, size)

        new_values = self._values.transpose(axis_indices)
        new_values = new_values.reshape(axis_sizes)

        new_axis_values = list()
        indices = np.zeros(count)

        for pos in range(size):
            current_values = [array[indices[k]] for k, array in enumerate(array_list)]
            new_axis_values.append(format.format(*current_values))

            # increment indices
            i = 0
            while True:
                indices[i] += 1
                if indices[i] < len(axes[i]):
                    break
                indices[i] = 0
                i += 1
                if i == count:
                    break

        new_axis = Index(new_axis_name, new_axis_values)
        new_axes.insert(0, new_axis)
        return self.__class__(new_axes,new_values)


    def change_axis(self, oldIndex, newIndex, compareMode=1,dtype=None, default=None):
        """Returns a new Cube instance with the axes changed,lookin by value o by position
        """
        if compareMode==1:

            if default is None:
                return self.__getitem__( oldIndex==newIndex , False )
                #return self[oldIndex==newIndex]
                
            else:
                defaultType = int
                if not default is None: 
                    defaultType = type(default)

                tmp_index = oldIndex == newIndex
                _m = None
                if len(self.values)>0 and len(tmp_index.values)>0:
                    _m=apply_op(self,tmp_index, np.vectorize(vectormul))
                else:
                    _m=apply_op(self,tmp_index, np.vectorize(vectormul,otypes=[defaultType]))

                
                _aux = _m.reduce(safesum, axis=oldIndex)
                
                if not default is None:
                    _def = ~tmp_index.any(axis=oldIndex)
                    _def = apply_op(_def,default, np.vectorize(vectormul))
                    _aux = apply_op(_aux,_def, np.vectorize(vectoradd))
                    return _aux
                
                if dtype is None:
                    return _aux
                else:
                    return self.__class__(_aux.axes,_aux.values.real.astype(dtype,copy=False))

        else:
            return self.replace_axis(oldIndex, newIndex)
 

    # ***************************************
    # *** Filtering, indexing and slicing ***
    # ***************************************

    def tryFilter(self, filter_by, values=None):
        """Returns a new Cube instance with filtered axes. if filter_by index not exists, return self cube
        """
        indexName = filter_by
        if isinstance(indexName,Index):
            indexName = indexName.name
            
        if is_axis(filter_by) and self.has_axis(indexName):
            return self.filter(filter_by, values)
        else:
            return self

    def filter(self, filter_by, values=None):
        """Returns a new Cube instance with filtered axes.
        :param filter_by: axis name (str), axis index (int), Axis, Cube or collection of Axis or Cube instances
            If filter_by is Cube or collection of Axis or Cube instances, then unmatched axes are ignored.
            If filter_by is axis name, index or Axis instance, then exception is raised if the axis cannot be matched.
        :param values: collection of values to be filtered; defined only if filter_by is str or int
        :return: new Cube instance
        """
        if isinstance(filter_by, str) or isinstance(filter_by, int):
            return self._filter_by_values(filter_by, values)

        if values is not None:
            raise ValueError("'values' can be non-None only when filtering by axis name or index")

        if is_axis(filter_by):
            if hasattr(filter_by, "__contains__"):
                # we intentionally do not pass axis.values because
                # the axis has (likely optimized) 'in' operator
                return self._filter_by_values(filter_by.name, filter_by)
            else:
                # else we provide raw values, but then the lookup is slower
                return self._filter_by_values(filter_by.name, filter_by.values)

        if hasattr(filter_by, "axes"):  # for cube-like objects
            filter_by = filter_by.axes

        # a collection of axes or cubes is expected
        result = self
        for item in filter_by:
            if not is_axis(item) or self.has_axis(item.name):  # skip unmatched axes
                result = result.filter(item)
        return result

    def exclude(self, axis, values):
        """Remove slices from cube which correspond to given values on an axis.
        :param axis: axis on which the values are to be removed
        :param values: values to remove
        :return: new Cube instance
        Note: Values which do not exist on the given axis are ignored. I.e. no error is raised.
        """
        axis, axis_index = self._axis_and_index(axis)
        value_indices = [i for i, v in enumerate(axis.values) if v not in values]
        return self.take(axis_index, value_indices)

    def take(self, axis, indices):
        """Filters the cube along an axis using specified indices. 
        Analogy to numpy.ndarray.take.
        :param indices: a collection of ints or int or slice
        :param axis: axis name (str), index (int) or instance
        :return: new Cube instance
        :raise LookupError: is the axis does not exist, ValueError for invalid indices
        If 'indices' is a single int, then the axis is removed from the cube.
        If 'indices' is a collection of ints, then the axis is preserved.
        """
        axis, axis_index = self._axis_and_index(axis)
        new_axis = axis.take(indices)
        if isinstance(indices, int):
            # if indices is a single int,
            # then will remove one dimension
            axes = self._axes.remove(axis_index)
        else:
            # otherwise the dimension is preserved,
            # even if the collection has one element
            axes = self._axes.replace(axis_index, new_axis)
        values = self._values.take(indices, axis_index)
        return self.__class__(axes,values)

    def slice(self, axis, *args):
        """Return sliced cube, the arguments have the same meaning as in standard slice() function.
        :param axis: axis name (str), index (int) or instance
        :param args: slice object or one to three parameters to create a slice object
        :return: new Cube instance

        Examples:
        c.slice(ax, 1)  # first item
        c.slice(ax, -1, None)  # last item
        c.slice(ax, 1, None)  # all except first item
        c.slice(ax, 0, -1)  # all except last item
        c.slice(ax, None, None, 2)  # every even item
        c.slice(ax, 1, None, 2)  # every odd item
        c.slice(ax, None, None, -1)  # reversed items
        """
        axis, axis_index = self._axis_and_index(axis)

        if len(args) == 1 and isinstance(args[0], slice):
            slc = args[0]
        else:
            slc = slice(*args)

        slices = [slice(None)] * self.ndim
        slices[axis_index] = slc
        new_axis = axis[slc]
        new_axes = self._axes.replace(axis_index, new_axis)
        return self.__class__(new_axes,self.values[slices])

    def first(self, axis, n=1):
        """Returns the subsection of the cube which corresponds to the first n values on the specified axis.
        :param axis: axis name (str), index (int) or instance
        :param n: number of values along the axis
        :return: new Cube instance
        """
        return self.slice(axis, 0, n)

    def last(self, axis, n=1):
        """Returns the subsection of the cube which corresponds to the last n values on the specified axis.
        :param axis: axis name (str), index (int) or instance
        :param n: number of values along the axis
        :return: new Cube instance
        """
        return self.slice(axis, -n, None)

    def reversed(self, axis):
        """Reverse the order of values along the axis.
        :param axis: axis name (str), index (int) or instance
        :return: new Cube instance
        """
        return self.slice(axis, None, None, -1)

    def compress(self, axis, condition):
        """Filters the cube along an axis using a boolean mask along a specified axis. 
        Analogy to numpy.ndarray.compress.
        :param axis: axis name (str), index (int) or instance
        :param condition: collection of boolean values
        :return: new Cube instance
        :raise LookupError: is the axis does not exist, # TODO - error if wrong type
        """
        axis, axis_index = self._axis_and_index(axis)
        new_axis = axis.compress(condition)
        axes = self._axes.replace(axis_index, new_axis)
        values = self._values.compress(condition, axis_index)
        return self.__class__(axes,values)

    # **************************
    # *** Data manipulations ***
    # **************************

    def set_data(self,axes,value):
        """ set cell data filtered by axes """
        slices = [slice(None)] * self.ndim
        for param_axis in axes:
            if self.has_axis(param_axis.name):
                axis, axis_index = self._axis_and_index(param_axis.name)
                wh = np.where(axis.values==param_axis.values)[0]
                if (len(wh)>0):
                    slices[axis_index] = int(wh[0])

        if isinstance(value,str):
            if kindToString(self.values.dtype.kind) == "string":
                _size = self.values.dtype.itemsize
                if self.values.dtype.char=="U":
                    _size/=4
                if len(value)>_size:
                    self._values = self.values.astype("U"+str(len(value)))

        self.values[tuple(slices)] = value


    def astype(self,dtype):
        """ change dtype of values of cube """
        return self.__class__(self.axes,self.values.real.astype(dtype,copy=False))



    # *********************************
    # *** Cube generating functions ***
    # *********************************
        

    def generateDefinition(self):
        """Generate code for cube definition"""
        indexes = str([idx.name for idx in self.axes]).replace("'",'')
        np.set_printoptions(threshold = np.prod(self.values.shape))
        data = np.array2string(self.values, separator=",").replace('\n','')
        if kindToString(self.values.dtype.kind)=="string" or kindToString(self.values.dtype.kind)=="object":
            deff = "result = cp.cube(" + indexes + "," + data + ", dtype='O')"
        else:
            deff = "result = cp.cube(" + indexes + "," + data + ")"
        return deff

    @staticmethod
    def full(axes, fill_value, dtype=None):
        """Returns a new cube filled with a uniform value.
        :param axes: a collection of Axis instances for form the new cube
        :param fill_value: the uniform value to fill the cube
        :param dtype: the value type of the new cube (usually int or float)
        :returns: new Cube instance
        """
        axes = make_axes(axes)
        shape = tuple(len(axis) for axis in axes)
        values = np.full(shape, fill_value, dtype)
        return Cube(axes,values)
        
    @staticmethod
    def zeros(axes, dtype=float):
        """Returns a new cube filled with zeros.
        :param axes: a collection of Axis instances for form the new cube
        :param dtype: the value type of the new cube (usually int or float)
        :returns: new Cube instance
        """
        axes = make_axes(axes)
        shape = tuple(len(axis) for axis in axes)
        values = np.zeros(shape, dtype)
        return Cube(axes,values )

    @staticmethod
    def ones(axes, dtype=float):
        """Returns a new cube filled with ones.
        :param axes: a collection of Axis instances for form the new cube
        :param dtype: the value type of the new cube (usually int or float)
        :returns: new Cube instance
        """
        axes = make_axes(axes)
        shape = tuple(len(axis) for axis in axes)
        values = np.ones(shape, dtype)
        return Cube(axes,values)

    @staticmethod
    def empty(axes, dtype=float):
        """Returns a new cube filled with random values.
        :param axes: a collection of Axis instances for form the new cube
        :param dtype: the value type of the new cube (usually int or float)
        :returns: new Cube instance
        """
        axes = make_axes(axes)
        shape = tuple(len(axis) for axis in axes)
        values = np.empty(shape, dtype)
        return Cube(axes,values)

    # *********************************
    # *** Private utility functions ***
    # *********************************

    def _axis_and_index(self, axis_id):
        return self._axes.axis_and_index(axis_id)

    def _filter_by_values(self, axis, values):
        """Returns a cube filtered by specified values on a given axis. Takes into account only values
        which exist on the axis. Other values are ignored.
        :param axis: axis name (str), index (int) or instance
        :param values: a collection of values providing 'in' operator
        :return: new Cube instance
        """
        axis, axis_index = self._axis_and_index(axis)
        value_indices = [i for i, v in enumerate(axis.values) if v in values]
        return self.take(axis_index, value_indices)

    def _align_axis(self, new_axis):
        """Returns a cube with values aligned to a new axis. The axis to be aligned has the same name as the new
        axis. The order of the axes in the cube remains the same. The new axis will become one of the cube axes.
        :param new_axis: Axis instance
        :return: new Cube instance
        :raise LookupError if new_axis cannot be matched to any axis in the cube.
        """
        old_axis, old_axis_index = self._axis_and_index(new_axis.name)
        indices = old_axis.indexof(new_axis.values)
        new_values = self._values.take(indices, old_axis_index)
        new_axes = self._axes.replace(old_axis_index, new_axis)
        return self.__class__(new_axes,new_values)

    def _aggregate(self, func, new_axes, axis_indices_to_remove):
        # new_axes - collection of axes in the result
        # axis_indices_to_remove - which axes should be removed by the aggregation
        new_values = self._values
        if axis_indices_to_remove:
            new_values = func(new_values, axis_indices_to_remove)
        if len(new_axes)==0:
            if isinstance(new_values,str):
                return new_values
            else:
                return new_values.item()
        else:
            return self.__class__(new_axes,new_values)

    def _group(self, axis, func, sorted=True, *args):  # **kwargs): # since numpy 1.9
        # Group the same values along a given axis by applying a function.
        # :param axis: name (str) or index (int) of axis to group the cube values by
        # :param func: aggregation function, e.g. np.sum, np.mean etc.
        #    There are the following requirements:
        #    - the function takes two fixed arguments - array and axis (given by index)
        #    - these two fixed arguments can be followed by a variable number of other arguments passed in *args
        #    - the function must return an array with one axis less then the input array
        old_axis, old_axis_index = self._axis_and_index(axis)

        # shortcut evaluation
        if isinstance(old_axis, Index):
            return self

        sub_cubes = list()

        if sorted:
            # np.unique sorts the returned values by default
            unique_values = np.unique(old_axis.values)
        else:
            # special handling is required if the first occurrence order is to be kept
            unique_values, unique_indices = np.unique(old_axis.values, return_index=True)
            index_array = np.argsort(unique_indices)
            unique_values = unique_values[index_array]

        old_values = old_axis.values
        all_indices = np.arange(len(old_values))
        for value in unique_values:
            indices = all_indices[old_values == value]
            sub_cube = self._values.take(indices, old_axis_index)
            sub_cube = np.apply_along_axis(func, old_axis_index, sub_cube, *args)  # , **kwargs) # since numpy 1.9
            sub_cube = np.expand_dims(sub_cube, old_axis_index)
            sub_cubes.append(sub_cube)

        # the created axis is Index because it has unique values
        new_axis = Index(old_axis.name, unique_values)
        new_axes = self._axes.replace(old_axis_index, new_axis)
        new_values = np.concatenate(sub_cubes, old_axis_index)
        return self.__class__(new_axes,new_values)


    def selectOptimalFunction(self, partA, partB, fn):
        res = fn

        if fn is np.add:
            if isinstance(partA,self.__class__) and isinstance(partB,self.__class__):
                if partA.values.dtype.kind in {'U', 'S'} or partB.values.dtype.kind in {'U', 'S'}:
                    res = np.vectorize(vectoradd)
            else:
                if isinstance(partB,str):
                    res = np.vectorize(vectoradd)
        elif fn is np.multiply:
            if isinstance(partA,self.__class__) and isinstance(partB,self.__class__):
                if partA.values.dtype.kind in {'U', 'S'} and partB.values.dtype.kind in {'b'} or partA.values.dtype.kind in {'b'} and partB.values.dtype.kind in {'U', 'S'}:
                    res = np.vectorize(vectormul)
            else:
                if isinstance(partB,bool):
                    res = np.vectorize(vectormul)

        return res


def apply_op(a, b, func, *args):
    """Apply function element-wise on values of two cubes.
    The cube axes are matched and aligned before the function is applied.
    :param a: Cube instance
    :param b: Cube instance
    :param func: function to be applied
    :param args: additional arguments which are passed to the function
    :return: new Cube instance
    """
    if is_axis(a):
        a = Cube([a],a.values)
    if is_axis(b):
        b = Cube([b],b.values)

    if not is_cube(a):
        _tmpValues = None
        try:
            _tmpValues = func(a, b.values, *args)
        except (ValueError, TypeError):
            _tmpValues = func(a, b.values.astype("O"), *args)


        if isinstance(_tmpValues,type(NotImplemented)):
            _tmpValues = _convertNpFunction(func)(a, b.values, *args)
        return Cube(tuple(b.axes),_tmpValues )


    if not is_cube(b):
        _tmpValues = None
        try:
            _tmpValues = func(a.values, b, *args)
        except (ValueError, TypeError):
            _tmpValues = func(a.values.astype("O"), b, *args)
        if isinstance(_tmpValues,type(NotImplemented)):
            _tmpValues =_convertNpFunction(func)(a.values, b, *args)
        return Cube(tuple(a.axes),_tmpValues)

    values_a = a.values
    values_b = b.values
    all_axes = list()

    for axis_index_a, axis_a in enumerate(a.axes):

        try:
            axis_b, axis_index_b = b._axis_and_index(axis_a.name)
        except LookupError:
            # axis not found in cube b --> do not align
            axis_b = axis_a

        # if axes are identical or if axis_b has not been found --> do not align
        if axis_b is axis_a:
            all_axes.append(axis_a)
            continue

        axis, values_a, values_b = align_arrays(axis_a, axis_b, axis_index_a, axis_index_b, values_a, values_b)
        all_axes.append(axis)

    # add axes from b which have not been aligned
    for axis_b in b.axes:
        if not a.has_axis(axis_b.name):
            all_axes.append(axis_b)

    values_a = broadcast_array(values_a, a._axes, all_axes)
    values_b = broadcast_array(values_b, b._axes, all_axes)
    
    try:
        _newValues = func(values_a, values_b, *args)
        if isinstance(_newValues,type(NotImplemented)):
            raise ValueError('NotImplemented')
        return Cube(all_axes,_newValues)
    except (ValueError, TypeError):
        newFun = np.ndarray.__eq__
        return Cube(all_axes,_convertNpFunction(func)(values_a, values_b, *args) )

def _convertNpFunction(fnSource):
    """ convert np function to np.ndarray function. Ex. np.equal to no.ndarray.__eq__"""
    fndic={
        "<ufunc 'equal'>": np.ndarray.__eq__,
        "<ufunc 'not_equal'>": np.ndarray.__ne__,
        "<ufunc 'less'>": np.ndarray.__lt__,
        "<ufunc 'less_equal'>": np.ndarray.__le__,
        "<ufunc 'greater'>": np.ndarray.__gt__,
        "<ufunc 'greater_equal'>": np.ndarray.__ge__
    }
    return fndic[str(fnSource)] if str(fnSource) in fndic else fnSource

def concatenate(cubes, axis_name, as_index=False, broadcast=False):
    """Joins cubes along one axis on which the cubes have non-overlapping values.
    :param cubes: a collection of Cube instances
    :param axis_name: the name of axis on which the cubes will be joined
    :param as_index: if True, the new joined axis will be created as Index; otherwise it will be Axis
    :param broadcast: allows automatic broadcasting of unique axes
    :return: new Cube instance
    :raise LookupError: if any cube does not contain the joined axis
    :raise ValueError: if Index instance shall be created but the values are not unique
    The joined axis becomes the first axis of the new cube regardless of its position in the original cubes.
    """

    main_axis_values_list = list()
    for cube in cubes:
        axis = cube.axis(axis_name)
        main_axis_values_list.append(axis.values)

    # concatenate the new main axis
    main_axis_values = np.concatenate(main_axis_values_list)
    if as_index:
        # will fail if does not have unique values
        main_axis = Index(axis_name, main_axis_values)
    else:
        main_axis = Axis(axis_name, main_axis_values)

    unique_axes_list = unique_axes_from_cubes(cubes)

    # create a unique list without the main axis
    unique_axes_list = [a for a in unique_axes_list if a.name != axis_name]

    return _align_broadcast_and_concatenate(cubes, unique_axes_list, main_axis, broadcast)


def stack(cubes, axis, broadcast=False):
    """Adds a new dimension and stack uniformly shaped cubes along this axis.
    This is different from concatenate which joins cubes along axis which already exists in all the cubes.
    :param cubes: a collection of Cube instances
    :param axis: Axis instance which is used to stack the cubes
    :param broadcast: allows automatic broadcasting of unique axes
    :return: new Cube instance with the new axis
    :raise ValueError: is an axis of the same axis name already exists in any of the cubes in the collection;
        ValueError if the axis has different length from the number of cubes in the collection
    """
    for cube in cubes:
        if cube.has_axis(axis.name):
            raise ValueError("cube already contains axis '{}'".format(axis.name))

    if len(cubes) != len(axis):
        raise ValueError("invalid axis length")

    unique_axes_list = unique_axes_from_cubes(cubes)

    return _align_broadcast_and_concatenate(cubes, unique_axes_list, axis, broadcast)


def _align_broadcast_and_concatenate(cube_list, axis_list, main_axis, broadcast):
    array_list = [cube.values for cube in cube_list]

    for base_axis in axis_list:
        for cube_index, cube in enumerate(cube_list):
            try:
                axis_index = cube.axis_index(base_axis.name)
            except LookupError:
                if broadcast:
                    continue
                else:
                    raise
            axis = cube.axis(axis_index)

            if axis is base_axis:
                # axes are identical, no need to align
                continue

            if is_indexed(axis):
                value_indices = axis.indexof(base_axis.values)
                array = array_list[cube_index]
                array_list[cube_index] = array.take(value_indices, axis_index)
            else:
                if not np.array_equal(axis.values, base_axis.values):
                    raise AxisAlignError("cannot align axes '{}' with unequal values".format(axis.name))

    # put the new main axis in front of the list
    axis_list.insert(0, main_axis)

    # broadcast value arrays
    for cube_index, cube in enumerate(cube_list):
        array = array_list[cube_index]
        array = broadcast_array(array, cube._axes, axis_list)
        array_list[cube_index] = array

    array_list = np.broadcast_arrays(*array_list)
    new_values = np.concatenate(array_list)
    return Cube(axis_list,new_values)


def is_cube(obj):
    return isinstance(obj, Cube)



def safeStringOpp(fn,a,i):
    try:
        if not a is None and isinstance(a,np.ndarray):
            if 0 in a.shape:
                return np.empty( (np.delete(np.array(a.shape),i)), dtype=a.dtype )
            if a.dtype.kind in {'U', 'S'}:
                return fn(a.astype('O'),axis=i)
        return fn(a,i)
    except:
        import pandas as pd
        return fn(pd.to_numeric(a.reshape(-1, ), errors='ignore').reshape(a.shape),axis=i) 


def safesum(a,i=None):
    return safeStringOpp(np.sum,a,i)
def safemean(a,i=None):
    return safeStringOpp(np.mean,a,i)
def safemax(a,i=None):   
    return safeStringOpp(np.max,a,i)
def safemin(a,i=None):
    return safeStringOpp(np.min,a,i)


def vectoradd(a,b): 
    if "numpy" in str(type(a)):
        a = a.item()
    if "numpy" in str(type(b)):
        b = b.item()
    return a+b

def vectorsub(a,b): 
    if "numpy" in str(type(a)):
        a = a.item()
    if "numpy" in str(type(b)):
        b = b.item()
    return a-b

def vectormul(a,b): 
    if "numpy" in str(type(a)):
        a = a.item()
    if "numpy" in str(type(b)):
        b = b.item()
    return a*b

def vectordiv(a,b): 
    if "numpy" in str(type(a)):
        a = a.item()
    if "numpy" in str(type(b)):
        b = b.item()
    return a/b


def kindToString(kind):
    if kind in {'U', 'S'}:
        return "string"
    elif kind in {'b'}:
        return "boolean"
    elif kind in {'i','u','f','c'}:
        return "numeric"
    elif kind in {'m','M'}:
        return "date"
    elif kind in {'O'}:
        return "object"
    elif kind in {'V'}:
        return "void"


import numpy as np
import pandas as pd
import xarray as xr
from pyplan_core import cubepy 
from pyplan_core.classes.evaluators.BaseEvaluator import BaseEvaluator
from pyplan_core.classes.evaluators.CubepyEvaluator import CubepyEvaluator
from pyplan_core.classes.evaluators.NumpyEvaluator import NumpyEvaluator
from pyplan_core.classes.evaluators.PandasEvaluator import PandasEvaluator
from pyplan_core.classes.evaluators.XArrayEvaluator import XArrayEvaluator
import inspect


class Evaluator(object):
    ipytonMethods = ["_repr_html_", "_repr_json_",
                     "_repr_jpeg_", "_repr_png_", "_repr_pretty_"]

    @staticmethod
    def createInstance(result):
        if result is None:
            return BaseEvaluator()
        else:
            if Evaluator.isPandas(result):
                return PandasEvaluator()
            elif Evaluator.isXArray(result):
                return XArrayEvaluator()
            elif Evaluator.isMatplotlib(result):
                from pyplan_core.classes.evaluators.MatplotlibEvaluator import MatplotlibEvaluator
                return MatplotlibEvaluator()
            elif Evaluator.isNumpy(result):
                return NumpyEvaluator()
            elif Evaluator.isBokeh(result):
                from pyplan_core.classes.evaluators.BokehEvaluator import BokehEvaluator
                return BokehEvaluator()
            elif Evaluator.isPlotly(result):
                from pyplan_core.classes.evaluators.PlotlyEvaluator import PlotlyEvaluator
                return PlotlyEvaluator()
            elif Evaluator.isCubepy(result):
                return CubepyEvaluator()
            elif Evaluator.isIPython(result):
                from pyplan_core.classes.evaluators.IPythonEvaluator import IPythonEvaluator
                return IPythonEvaluator()
            else:
                return BaseEvaluator()

    @staticmethod
    def isPandas(result):
        return isinstance(result, pd.DataFrame) or isinstance(result, pd.Series) or isinstance(result, pd.Index)

    @staticmethod
    def isXArray(result):
        return isinstance(result, xr.DataArray)

    @staticmethod
    def isMatplotlib(result):
        try:
            from matplotlib.artist import Artist as MatplotlibArtist
            return isinstance(result, MatplotlibArtist) or inspect.ismodule(result) and "matplotlib.pyplot" in str(result) or isinstance(result, np.ndarray) and result.ndim > 0 and len(result) > 0 and isinstance(result.item(0), MatplotlibArtist)
        except:
            return False

    @staticmethod
    def isNumpy(result):
        return isinstance(result, np.ndarray)

    @staticmethod
    def isBokeh(result):
        try:
            from bokeh.plotting import Figure
            from bokeh.layouts import LayoutDOM
            return isinstance(result, Figure) or isinstance(result, LayoutDOM)
        except:
            return False

    @staticmethod
    def isPlotly(result):
        try:
            from plotly.graph_objs._figure import Figure as PlotlyFigue
            return isinstance(result, PlotlyFigue)
        except:
            return False

    @staticmethod
    def isCubepy(result):
        return isinstance(result, cubepy.Cube) or isinstance(result, cubepy.Index)

    @staticmethod
    def isIPython(result):
        _dir = dir(result)
        return len(list(set(_dir) & set(Evaluator.ipytonMethods))) > 0



import pandas as pd
import xarray as xr
import numpy as np
import re
import os
import importlib
import subprocess
try:
    from StringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO


class PyplanFunctions(object):

    def __init__(self, model=None):
        self.model = model

    def release(self):
        self.model = None

    def set_domain(self, dataArray, domainDic, defaultValue=None):
        """ Reindex the dataArray by applying the indices of the domainDic param
            Ex. 
                pp.set_domain(da,{"time":time_idex, "products":product_index})
        """
        _da = dataArray
        for key in domainDic:
            _da = _da.reindex({key: domainDic[key].values})
            _da = _da.rename({key: domainDic[key].name})
        if not defaultValue is None:
            _da = _da.fillna(defaultValue)
        return _da

    def build_report(self, values, name="Report", report_index=None):
        """ DEPRECATED. Use the create_report function instead """
        _titles = [str(xx.name) for xx in values]
        _index = None
        if report_index is None:
            _index = pd.Index(_titles, name=name)
        else:
            _index = report_index

        return xr.concat(values, _index)

    def create_dataarray(self, value, coords, dtype=None):
        """ Creates a dataarray using an atomic value distributed along all dimensions 
            Ex. 
                pp.create_dataarray(1., coords=[time_idex, product_index])
        """
        _data = np.full(tuple([(len(x)) for x in coords]), value, dtype=dtype)
        return xr.DataArray(_data, coords)

    def find(self, param1, param2, compareType=1, caseSensitive=True):
        """
        param1: value or indexarray for compare
        param2: index compare to
        compareType: exact=1, start_with=2, end_with=3, contain=4  
        caseSensitive: able to differentiate between uppercase and lowercase (by default True)

        If param1 is a scalar (numeric or str) and param2 is an index:  return a dataArray indexed by param2 with True on ocurrences of param2
            Ex. pp.find("te", region, cp.end_with)
        If param1 is an index and param2 is an index too:  return a dataArray indexed by param1 and param2 with True on ocurrences of param1 on param2
            Ex. pp.find(subregion, region, cp.contain)

        """
        def _internalFn(item, value):
            if not isinstance(item, str):
                item = str(item)
            if not isinstance(value, str):
                value = str(value)

            if compareType == 1:
                if caseSensitive:
                    return item == value
                else:
                    return item.lower() == value.lower()
            elif compareType == 2:
                if caseSensitive:
                    return item[:len(value)] == value
                else:
                    return item[:len(value)].lower() == value.lower()
            elif compareType == 3:
                if caseSensitive:
                    return item[-len(value):] == value
                else:
                    return item[-len(value):].lower() == value.lower()
            elif compareType == 4:
                if caseSensitive:
                    return value in item
                else:
                    return value.lower() in item.lower()

        if (isinstance(param1, str) or str(param1).isnumeric()) and isinstance(param2, pd.Index):
            vfn = np.vectorize(_internalFn)
            return xr.DataArray(vfn(param2.values, param1), [param2])

        if isinstance(param1, pd.Index) and isinstance(param2, pd.Index):
            _res = self.create_dataarray(False, [param1, param2], dtype=bool)
            for row in param1.values:
                for col in param2.values:
                    _res.loc[{param1.name: slice(row, row), param2.name: slice(
                        col, col)}] = _internalFn(col, row)
            return _res

    def apply_fn(self, obj, applyFn, *args):
        """ Apply "applyFn" to "obj" where obj can be DataArray or Index
            Ex.
                pp.apply(dataArray, node_function)
        """
        vfn = np.vectorize(applyFn)
        if isinstance(obj, pd.Index):
            return pd.Index(np.unique(vfn(obj.values, *args)))
        if isinstance(obj, xr.DataArray):
            return xr.apply_ufunc(vfn, obj, *args)
        return None

    def subset(self, cube):
        """Returns an index with the elements of the index for which cube is true. The function is used to create a new index that is a subset of an existing index.
            Ex. pp.subset(sales>0)
        """
        cond = cube > 0
        values = cond.coords[cond.dims[0]].values[cond.values]
        return pd.Index(values)

    def split_text(self, param1, separator, part=None):
        """
            Description: Returns a DataArray object with text values formed by splitting the elements of param1 text values at each occurrence of separator "separator". 
            The DataArray will have the original dimension plus a new dimension 'Parts' of length (number of separators + 1). 
            All text values must have the same number of separators separator.        
        """
        if isinstance(param1, pd.Index):
            param1 = xr.DataArray(param1.values, [param1])

        _q_separators = self.apply_fn(param1, lambda x: x.count(separator))
        _max_q_separators = np.asscalar(_q_separators.max().values)
        _result_coords = ['Part ' + str(i)
                          for i in range(1, _max_q_separators + 2)]
        _result_dim = pd.Index(_result_coords)
        _result_dim.name = "Parts"

        _results = []

        for _part in range(_max_q_separators + 1):
            _dataarray = self.apply_fn(
                param1, lambda x: x.split(separator)[_part])
            _results.append(_dataarray)

        _res = xr.concat(_results, dim=_result_dim)
        if not part is None:
            _res = _res.sel(Parts="Part " + str(part), drop=True)

        return _res

    def get_pos(self, index):
        """
            Return datarray with pd.index and yours positions as values
            Ex. pp.get_pos(time_index)
        """
        return xr.DataArray(range(0, len(index)), [index])

    def concat_index(self, *args):
        """
        Concatenates two or more indexes and/or atomic values into a single new index
        Return: new index
            Ex.
                pp.concatIndex(index1,index2,index3,value1,value2)
        """
        _list = []
        for arg in args:
            if isinstance(arg, pd.Index):
                values = (arg.values).tolist()
                _list.extend(values)
            else:
                _list.append(arg)

        seripandas = pd.Series(_list)
        return pd.Index(seripandas.unique())

    def linear_depreciation(self, investments, usefulLife, timeIndex, includeInCurrentMonth=False, timeIndexFormat='%Y.%m'):
        """
            Returns the straight-line depreciation of dataArray investments over its usefulLife.

            investments: DataArray containing investments
            usefulLife: DataArray with number of years of life expectancy
            timeIndex: Time dimension of dataArray. Must be a Pandas Index
            includeInCurrentMonth: Wheter to start depreciating in month t or month t+1
            timeIndexFormat: i.e. for '2016.01' would be '%Y.%m'
        """
        # Depreciation amount (safe division by zero)
        _usefulLife_months = usefulLife.astype(int) * 12
        _usefulLife_months_den = xr.where(
            _usefulLife_months == 0, 1, _usefulLife_months)
        _depreciation = xr.where(_usefulLife_months ==
                                 0, 0, investments / _usefulLife_months_den)

        # Calculate first and last months to depreciate
        _df_per = _usefulLife_months.to_dataframe('first').reset_index()
        _df_per['key'] = 1
        _df_time = pd.DataFrame(timeIndex)
        _df_time['key'] = 1
        _df = _df_time.merge(_df_per, on=['key']).drop(columns=['key'])
        _df['ending'] = pd.to_datetime(
            _df[timeIndex.name].str.replace('.', '-')).dt.to_period('M')
        _df['ending'] = (_df['ending'] + _df['first']
                         ).dt.strftime(timeIndexFormat)

        # Get dimensions indexes and names
        _getNodeFn = self.model.getNode

        _da_dims_names = list(usefulLife.dims)
        _da_dims = {timeIndex.name: timeIndex}
        for dim in _da_dims_names:
            _da_dims.update({dim: _getNodeFn(dim).result})

        # DataArray with ending date
        _ending = self.dataarray_from_pandas(
            _df, _da_dims, valueColumns='ending', defaultValue='')

        # Allocate depreciation to corresponding periods
        _depreciations = investments * 0.
        for t in timeIndex:
            _ending_month = self.subscript(_ending, timeIndex, t)
            _depreciation_amount_t = self.subscript(
                _depreciation, timeIndex, t)
            if includeInCurrentMonth:
                _depreciacion_t = xr.where((self.to_dataarray(timeIndex) >= t) & (
                    self.to_dataarray(timeIndex) < _ending_month), _depreciation_amount_t, 0.)
            else:
                _depreciacion_t = xr.where((self.to_dataarray(timeIndex) > t) & (
                    self.to_dataarray(timeIndex) <= _ending_month), _depreciation_amount_t, 0.)
            _depreciations = _depreciations + _depreciacion_t

        return _depreciations

    def irr(self, flow, time_index):
        """Returns the Internal Rate of Return (IRR) of a series of periodic payments (negative values) and inflows (positive values). The IRR is the discount rate at which the Net Present Value (NPV) of the flows equals zero. 
            The variable flow must be indexed by time_index.

            If the cash flow never changes sign, cp.irr() has no solution and returns NAN (Not A Number).
        """
        _getNodeFn = self.model.getNode

        _rest_of_indexes_labels = np.setdiff1d(flow.dims, [time_index.name])

        _cube = None
        if len(_rest_of_indexes_labels) == 0:
            _cube = np.irr(flow.values)
        else:
            _rest_of_indexes = [_getNodeFn(
                xx).result for xx in _rest_of_indexes_labels]
            _cube = self.create_dataarray(0., _rest_of_indexes)
            _multivalues = [idx.values for idx in _rest_of_indexes]
            _values = pd.MultiIndex.from_product(_multivalues).values

            for _item in _values:
                _filter = {}
                for _nn in range(len(_item)):
                    _filter[_rest_of_indexes[_nn].name] = _item[_nn]
                _toIrr = flow.sel(_filter).values
                _irr = np.irr(_toIrr)
                _cube.loc[_filter] = _irr
        return _cube

    def copy_as_values(self, source, targetId):
        """Copy values of datArray "source" into dataArray with id 'targetId'. This function alter the definition of dataArray with 'targetId' identifier.
                source: dataArray/index from copy values
                targetId: identifier (string) of the target node 
        """
        _getNodeFn = self.model.getNode

        if isinstance(source, str):
            source = _getNodeFn(source).result

        if not isinstance(source, xr.DataArray) and not isinstance(source, pd.Index) and not isinstance(source, float) and not isinstance(source, int):
            raise ValueError(
                "The 'source' parameter must be a xr.DataArray, pd.Index, float or int")

        if not isinstance(targetId, str):
            raise ValueError(
                "The 'targetId' parameter must be a string (identifier of node)")

        newDef = ""
        if isinstance(source, float) or isinstance(source, int):
            newDef = f"result = {str(source)}"
        elif isinstance(source, xr.DataArray):
            _indexes = str(list(source.dims)).replace("'", '')
            np.set_printoptions(threshold=np.prod(source.values.shape))
            _data = np.array2string(source.values, separator=",", precision=20, formatter={
                                    'float_kind': lambda x: "np.nan" if np.isnan(x) else repr(x)}).replace('\n', '')
            newDef = f"result = xr.DataArray({_data},{_indexes})"
        elif isinstance(source, pd.Index):
            np.set_printoptions(threshold=np.prod(source.values.shape))
            _data = np.array2string(source.values, separator=",", precision=20, formatter={
                                    'float_kind': lambda x: "np.nan" if np.isnan(x) else repr(x)}).replace('\n', '')
            newDef = f"result = pd.Index({_data})"

        _getNodeFn(targetId).definition = newDef
        return True

    def excel_connection(self, filepath, useOpenpyxl=False, dataOnly=True, readOnly=True):
        """ Create excel object from filepath.
            filepath: path to excel file
            useOpenpyxl: True for use custom 
            dataOnly: True for view only the values, not formula
            readOnly: True for read only, False for write options
            Ex.
                pp.excel_connection("\\path\\to\\the\\excelfile.xlsx")
        """

        if self.model.isLinux():
            filepath = filepath.replace("\\", "/")

        _getNodeFn = self.model.getNode
        fullFilename = filepath

        if not os.path.isfile(fullFilename):
            fullFilename = _getNodeFn("current_path").result + filepath

        if os.path.isfile(fullFilename):
            if useOpenpyxl:
                from openpyxl import load_workbook
                return load_workbook(fullFilename, data_only=dataOnly, read_only=readOnly)
            else:
                return filepath
        else:
            raise ValueError("File not found")

    def subscript(self, dataArray, indexes, values):
        """
            Filter dataArray using the filterList filters. 

            dataArray: dataArray to be filtered
            indexes: the index to filter 
            values: the value to filter 
            Ex.
                pp.subscript(dataArray, index, value)
        """
        if not isinstance(dataArray, xr.DataArray):
            raise ValueError(
                "the 'dataArray' parameter must be of the type xr.DataArray")

        if not isinstance(indexes, list):
            indexes = [indexes]
        if not isinstance(values, list):
            values = [values]

        res = dataArray
        filterDic = {}
        for _pos, indexItem in enumerate(indexes):
            filterDic[indexItem.name] = values[_pos]

        if len(filterDic) > 0:
            res = res.sel(filterDic, drop=True)

        return res

    def change_index(self, dataArray, oldIndex, newIndex, compareMode=1, defaultValue=None):
        """ Change index of a dataArray object.
            compareMode: 1: by Value (default), 2: by pos
            Ex.
                pp.change_index(dataArray, oldIndex, newIndex)
        """
        _da = dataArray

        if compareMode == 1:
            _temp = _da.reindex({oldIndex.name: newIndex.values})
            _temp[newIndex.name] = _temp[oldIndex.name]
            _temp = _temp.swap_dims(
                {oldIndex.name: newIndex.name}).drop(oldIndex.name)
            if not defaultValue is None:
                _temp = _temp.fillna(defaultValue)
            return _temp
        else:

            if len(oldIndex.values) == len(newIndex.values):
                _tmp = _da.copy()
                _tmp.coords[newIndex.name] = _tmp.coords[oldIndex.name]
                _tmp.coords[newIndex.name].data = newIndex.values
                _tmp = _tmp.swap_dims(
                    {oldIndex.name: newIndex.name}).drop(oldIndex.name)
                return _tmp
            elif len(oldIndex.values) > len(newIndex.values):
                raise ValueError(
                    "Changeindex by pos for indices of different size is not implemented")
            else:
                raise ValueError(
                    "Changeindex by pos for indices of different size is not implemented")

    def kind_to_string(self, kind):
        """Returns the data type on human-readable string
        """
        if kind in {'U', 'S'}:
            return "string"
        elif kind in {'b'}:
            return "boolean"
        elif kind in {'i', 'u', 'f', 'c'}:
            return "numeric"
        elif kind in {'m', 'M'}:
            return "date"
        elif kind in {'O'}:
            return "object"
        elif kind in {'V'}:
            return "void"

    def pandas_from_excel(self, excel, sheetName=None, namedRange=None, cellRange=None, indexes=None, driver="Driver={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)};DBQ=%s;READONLY=TRUE"):
        """ Return a pandas dataframe from excel.
            excel: path to excel file or cp.excel object
            sheetName: sheet name to be read
            namedRange: name of the range to be read
            cellRange: used with sheetname, for read from a simple range
            indexes: Listo of columns names for convert to index of dataframe
            Ex.
                pp.pandas_from_excel(excelNode,"Sheet 1")
                pp.pandas_from_excel(excelNode,namedRange="name_range")
                pp.pandas_from_excel(excelNode,"Sheet 1",cellRange="A1:H10")
        """
        if isinstance(excel, str):
            if not os.path.isfile(excel):
                excel = os.path.join(self.model.getNode(
                    "current_path").result, excel)

            filename = excel
            target_dir, single_filename = os.path.split(filename)
            file_name, _ = os.path.splitext(single_filename)
            target_dir = os.path.join(target_dir, file_name)

            file_to_read_legacy = os.path.join(
                target_dir, (f"{namedRange if namedRange else ''}.pkl"))
            target_dir = f"{target_dir[:target_dir.rfind(os.path.sep)+1]}.{target_dir[target_dir.rfind(os.path.sep)+1:]}"
            file_to_read = os.path.join(target_dir, f"{namedRange if namedRange else ''}.pkl") if os.path.isfile(
                os.path.join(target_dir, f"{namedRange if namedRange else ''}.pkl")) else file_to_read_legacy

            if os.path.isfile(file_to_read):
                df = pd.read_pickle(file_to_read, compression='gzip')
                if not indexes is None:
                    df.set_index(indexes, inplace=True)
                return df
            else:
                from openpyxl import load_workbook
                _wb = load_workbook(filename, data_only=True, read_only=True)
                return self.pandas_from_excel(_wb, sheetName, namedRange, cellRange, indexes)
        else:
            if "openpyxl.workbook" in str(type(excel)):
                rangeToRead = None
                if not namedRange is None:
                    the_range = excel.defined_names[namedRange]
                    dests = the_range.destinations
                    for title, coord in dests:
                        ws = excel[title]
                        rangeToRead = ws[coord]
                elif not cellRange is None:
                    ws = excel[sheetName]
                    rangeToRead = ws[cellRange]
                else:
                    rangeToRead = excel[sheetName]
                nn = 0
                cols = []
                values = []
                for row in rangeToRead:
                    if nn == 0:
                        cols = [str(c.value) for c in row]
                    else:
                        values.append([c.value for c in row])
                    nn += 1
                df = pd.DataFrame(values, None, cols)
                if not indexes is None:
                    if isinstance(indexes, str):
                        indexes = [indexes]
                    toIndex = []
                    for indexColumn in indexes:
                        if indexColumn in df.columns.values:
                            toIndex.append(indexColumn)
                    if len(toIndex) > 0:
                        df.set_index(toIndex, inplace=True)
                return df.dropna(how="all")
            else:
                raise ValueError("excel can be cp.excel object")

    def index_from_pandas(self, dataframe, columnName=None, removeEmpty=True):
        """ Return a pd.Index from an column of a pandas dataframe.
            dataframe: pandas dataframe
            columnName: dataframe column name used for create cp.index. By default is created using the first column
            removeEmpty: True for remove empty rows
            Ex.
                pp.index_from_pandas(df)
                pp.index_from_pandas(df,"column10")
        """
        _serie = None
        if columnName is None:
            _serie = dataframe[dataframe.columns[0]]
        else:
            _serie = dataframe[columnName]

        if removeEmpty:
            _serie.dropna(inplace=True)
            if self.kind_to_string(_serie.dtype.kind) == "string" or self.kind_to_string(_serie.dtype.kind) == "object":
                _serie = _serie[_serie != ""]

        return pd.Index(_serie.unique())

    def index_from_excel(self, excel, sheetName=None, namedRange=None, cellRange=None, columnName=None, removeEmpty=True):
        """ Return a pd.Index from an excel file.
            excel: pp.excel object
            sheetName: sheet name to be read
            namedRange: name of the range to be read
            cellRange: used with sheetname, for read from a simple range
            columnName: dataframe column name used for create pp.index. By default is created using the first column
            removeEmpty: True for remove empty rows
            Ex.
                pp.index_from_excel(excelNode,"Sheet 1")
                pp.index_from_excel(excelNode,namedRange="name_range")
                pp.index_from_excel(excelNode,namedRange="name_range", columnName="indicadores")
        """
        if isinstance(excel, str) or "openpyxl.workbook" in str(type(excel)):
            _df = self.pandas_from_excel(
                excel, sheetName, namedRange, cellRange)
            return self.index_from_pandas(_df, columnName, removeEmpty)
        else:
            raise ValueError(
                "excel can be excel_connection object or a str path to the filename")

    def dataarray_from_pandas(self, dataframe, domainDic, valueColumns, defaultValue=None, valueColumnsAsDim=True, sumDuplicateRecords=True):
        """ Returns a DataArray (valueColumns is string or (valueColumns is pd.Index and valueColumnsAsDim is True)) or Dataset (valueColumns is a list or (valueColumns is a pd.Index and valueColumnsAsDim is False)) from a Pandas dataframe applying the set_domain function.
            dataframe: Pandas dataframe with no index columns.
            domainDic: Dictionary of column names and index names. Ex. {'Column Name': index_name}.
            valueColumns: String, list or pd.Index. Dataframe's value columns.
            defaultValue: Default value when applying set_domain function.
            valueColumnsAsDim: If True, valueColumns becomes a dimension of resulting DataArray. If False, each value column becomes a variable of the resulting Dataset.
            sumDuplicateRecords: If True, sums identical rows. Otherwise, removes duplicates (except the first one). 

            Ex. 
                pp.dataarray_from_pandas(sales_dataframe, {'Sales Channel': sales_channels, 'Market Segment': market_segments, 'Month': time}, 'Sales', 0.)
        """
        _index_value_columns = None

        # Check if valueColumns is string, list, np.ndarray or pd.Index (transform to list) and indexes is dict.
        if isinstance(valueColumns, pd.Index):
            _index_value_columns = valueColumns.copy()
            _index_value_columns_name = _index_value_columns.name
            valueColumns = valueColumns.values.tolist()
        elif isinstance(valueColumns, np.ndarray):
            valueColumns = valueColumns.tolist()
        elif not isinstance(valueColumns, str) and not isinstance(valueColumns, list):
            raise ValueError(
                "valueColumns must be a string, a list or a pd.Index")
        if not isinstance(domainDic, dict):
            raise ValueError("Indexes must be a dictionary")

        # Transform indexes into list and create list with all columns.
        _index_cols = list(domainDic.keys())
        _cols = _index_cols.copy()

        if isinstance(valueColumns, list):
            _cols = _cols + valueColumns
        else:
            _cols.append(valueColumns)

        # If valueColumnsAsDim is True, check if every column is in dataframe and filter it.
        if (valueColumnsAsDim is True) and isinstance(_index_value_columns, pd.Index):
            _df_columns = dataframe.columns.values.tolist()
            _cols = [value for value in _df_columns if value in _cols]
            _filtered_value_columns = [
                value for value in _cols if value not in _index_cols]

        # Filter dataframe by columns.
        _df = dataframe[_cols]

        # Sum identical rows or remove duplicates.
        if sumDuplicateRecords is True:
            _df = _df.groupby(_index_cols, as_index=False).sum()
        else:
            _duplicate_rows = _df.duplicated(_index_cols)
            _df = _df[~_duplicate_rows]

        # If valueColumnsAsDim is True, melt valueColumns.
        if (valueColumnsAsDim is True) and isinstance(_index_value_columns, pd.Index):
            # Unpivot dataframe from wide format to long format by valueColumns.
            _df = pd.melt(_df, id_vars=_index_cols, value_vars=_filtered_value_columns,
                          var_name=_index_value_columns_name, value_name='values')
            _index_cols = _index_cols + [_index_value_columns_name]
            domainDic[_index_value_columns_name] = _index_value_columns

            # Create DataArray
            _data = _df.set_index(_index_cols)['values'].to_xarray()

            # Appy set_domain function to DataArray / Dataset.
            _data = self.set_domain(_data, domainDic, defaultValue)
        else:
            # Create DataArray / Dataset.
            _data = _df.set_index(_index_cols)[valueColumns].to_xarray()

            # Appy set_domain function to DataArray / Dataset.
            _data = self.set_domain(_data, domainDic, defaultValue)

        return _data

    def dataarray_from_excel(self, excel, sheetName=None, namedRange=None, cellRange=None, indexes=None, valueColumns=None, indexColumnHeaders=None, replaceByIndex=None, defaultValue=0):
        """ Return a xr.DataArray from excel file.
            excel: excel_connection object.
            sheetName: sheet name to be read
            namedRange: name of the range to be read.
            cellRange: used with sheetName to read from a simple range.
            indexes: pd.Index objects to perform a change_index operation.
            valueColumns: string with the column name of the dataframe that contains the values.
                        pd.Index with column names to convert columns to index.
            indexColumnHeaders: (optional) column names in pandas to parse with indexes. Used if header on dataframe is not equal to index identifiers.
            replaceByIndex: (optional) replace index used in valueColumns by this index (using change_index).

            Ex.
                pp.dataarray_from_excel(excelNode,"Sheet 1",indexes=[indicadores],valueColumns="descuentos")
                pp.dataarray_from_excel(excelNode,namedRange="nombre_rango",indexes=[indicadores],valueColumns=time)
        """

        dataframe = self.pandas_from_excel(
            excel, sheetName, namedRange, cellRange)
        # Check size of dataframe. If it is empty, create empty dataArray. Else, proceed
        if len(dataframe) == 0:
            if not isinstance(indexes, list):
                indexes = [indexes]
            if isinstance(valueColumns, pd.Index):
                indexes.append(valueColumns)

            _data = np.full(tuple([(len(x)) for x in indexes]), defaultValue)
            return xr.DataArray(_data, indexes)
        else:
            valueIndex = None
            if isinstance(valueColumns, pd.Index):
                valueIndex = valueColumns
                valueColumns = valueIndex.values
            elif isinstance(valueColumns, str):
                valueColumns = [valueColumns]

            if indexColumnHeaders is None:
                indexColumnHeaders = [index.name for index in indexes]

            # Create total index and index names
            _allindexes = indexes
            _allIndexNames = indexColumnHeaders[:]
            if not valueIndex is None:
                _allindexes.append(valueIndex)
                _allIndexNames.append("data_index")

            # fill other columns for prevent melt error
            cols_not_in_df = [
                col for col in valueColumns if col not in dataframe.columns]
            for col in cols_not_in_df:
                dataframe[col] = np.nan

            _full = dataframe.reset_index().melt(id_vars=indexColumnHeaders,
                                                 value_vars=valueColumns, var_name="data_index", value_name="data_value")

            # sum for acum over duplicate records
            _full = _full.groupby(_allIndexNames, as_index=False).sum()
            _dtype = _full["data_value"].dtype

            _dataType = self.kind_to_string(_dtype.kind)
            if _dataType == "string":
                _full = _full[(_full["data_value"] != "") &
                              (_full['data_value'].notna())]
            else:
                _full = _full[(_full["data_value"] != 0) &
                              (_full['data_value'].notna())]

            _full.set_index(_allIndexNames, inplace=True)
            _da = _full["data_value"].to_xarray()

            # If indexed, rename index
            if not indexes is None and not indexColumnHeaders is None:
                if not isinstance(indexes, list):
                    indexes = [indexes]
                idxPos = 0
                for cubeIndex in indexes:
                    newIndexName = cubeIndex.name
                    if idxPos <= len(indexColumnHeaders)-1:
                        oldIndexName = indexColumnHeaders[idxPos]
                        if not newIndexName in _da.coords:
                            _da.coords[newIndexName] = _da.coords[oldIndexName]
                            _da = _da.swap_dims(
                                {oldIndexName: newIndexName}).drop(oldIndexName)
                        idxPos += 1
                        # Reindex to complete combinations
                        _da = _da.reindex({newIndexName: cubeIndex.values})

            if not valueIndex is None:
                newIndexName = valueIndex.name
                oldIndexName = "data_index"
                if not newIndexName in _da.coords:
                    _da.coords[newIndexName] = _da.coords[oldIndexName]
                    _da = _da.swap_dims(
                        {oldIndexName: newIndexName}).drop(oldIndexName)
                # Reindex to complete combinations
                _da = _da.reindex({newIndexName: valueIndex.values})

                if not replaceByIndex is None:
                    _da = self.change_index(_da, valueIndex, replaceByIndex, 2)

            return _da.fillna(defaultValue)

    def to_dataarray(self, index):
        """ Convert an index into dataarray indexed by it and it as values
            Ex.
                pp.to_dataarray(time_index)
        """
        return xr.DataArray(index.values, [index])

    def goal_seek(self, nodeIdX, nodeIdObjective, goal=0, startValue=1, matrixIndex=None):
        """ Finds the value of nodeIdX that makes nodeIdObjective equal to goal.
            nodeIdX: String with id of node X
            nodeIdObjective: String with id of node X
            matrixIndex: Index for multidimensional goal seek
        """
        _getNodeFn = self.model.getNode
        if self._exists_module("scipy"):
            from scipy.optimize import newton

            if matrixIndex is None:
                def _f(x):
                    _getNodeFn(nodeIdX).definition = "result = " + str(x)
                    value = _getNodeFn(nodeIdObjective).result
                    return value - goal
                _res = newton(_f, x0=startValue)
                return _res
            else:
                raise ValueError(
                    "Multidimensional goal seek has not Implemented")
        else:
            raise ValueError("scipy library not found")

    def _exists_module(self, import_name):
        """Return true if module is intalled
        """
        try:
            importlib.import_module(import_name)
            return True
        except ImportError:
            return False

    def install_library(self, pypi_name, import_name=None):
        """ DEPRECATED. Use Lib manager instead
        """
        if import_name is None:
            import_name = pypi_name

        if not self._exists_module(import_name):
            # check in lib folder
            # install lib
            os.system(f"pip install {pypi_name}")
            importlib.invalidate_caches()
            if not self._exists_module(import_name):
                raise ValueError(f"Can't install the module '{import_name}'")
        return True

    def create_time(self, date_start, date_end, freq='M', format='%Y.%m'):
        """Create time index usign start and end dates and freq. The result is formated to format parameter
            Ex.
                pp.create_time('2016.01','2018.12')
                pp.create_time('2016.01.01','2016.12.31',freq='D',format='%d/%m/%Y')
        """
        if "." in date_start:
            date_start = date_start.replace('.', '-')
        if "." in date_end:
            date_end = date_end.replace('.', '-')
        return pd.Index(pd.period_range(start=date_start, end=date_end, freq=freq).strftime(format))

    def lookup(self, dataArray, dataMap, sharedIndex, defaultValue=0):
        """Returns the value of dataArray indexed by the index of dataMap.
            dataArray must be indexed by sharedIndex and dataArray values must correspond to elements of sharedIndex.
            For example: Let's say you have a cube with an estimated inflation rate by Country ("inflation_rate" is the name of the cube; "country" is the name of the index) and you want to assign it to the corresponding Company depending on its location. On the other hand, there's a many-to-one map where each Company is allocated to a single Country ("country_to_company_allocation"). The sharedIndex, in this case, is Country ("country").
            As a result, 
                lookup( inflation_rate , country_to_company_allocation , country )
            will return the estimated inflation rate by Company.
        """

        try:
            return dataArray.sel({sharedIndex.name: dataMap}, drop=True)
        except:
            valuesOk = dataMap[dataMap.isin(sharedIndex.values)]
            lookOk = dataArray.sel({sharedIndex.name: valuesOk}, drop=True)
            final = lookOk.reindex(
                {dataMap.dims[0]: dataMap.coords[dataMap.dims[0]].values})
            return final.fillna(defaultValue)

    def aggregate(self, dataArray, mapInfo, sourceIndex, targetIndex, aggregationFunction='sum'):
        """ Converts dataArray, originally indexed by sourceIndex, to a dataArray indexed by targetIndex, aggregating according to the mapInfo‘s allocation of targetIndex: sourceIndex.
            mapInfo gives the value of targetIndex for each element of sourceIndex (If the map does not match then the element will not be set into target index and information will be lost)
            aggregationFuction (optional) especifies the function to be used when grouping data (sum, mean, min, max, median)

            Ex. for aggregating time information into annual index, the syntax is:
                pp.aggregate(dataArray, timeToYearsMap, time, years)
        """
        # Transform map and targetIndex to list
        if not isinstance(mapInfo, list):
            mapInfo = [mapInfo]
        if not isinstance(targetIndex, list):
            targetIndex = [targetIndex]

        if len(mapInfo) == len(targetIndex):
            # Create dataframe map with new indexes
            _map = pd.DataFrame(columns=[sourceIndex.name]).set_index(
                sourceIndex.name)
            for i in range(len(mapInfo)):
                _map_i = mapInfo[i].to_dataframe(targetIndex[i].name)
                _map = _map.join(_map_i, how='outer')

            _df = dataArray.to_dataframe('value')
            _empty_filter = _df["value"] != 0
            # Drop rows with 0 if dataframe is not empty (to avoid error)
            if len(_df[_empty_filter]) != 0:
                _df = _df[_empty_filter]

            # Join new dimensions (target) and drop original (source)
            _df = _df.join(_map).reset_index()
            _df.drop(columns=[sourceIndex.name], inplace=True)
            _newDimList = [
                xx for xx in dataArray.dims if xx not in [sourceIndex.name]]

            # Groupby dataframe by new dimensions
            for i in range(len(targetIndex)):
                _newDimList.append(targetIndex[i].name)
            _df = _df.groupby(_newDimList).agg(aggregationFunction)

            # Transform to Xarray DataArray
            _da = _df["value"].to_xarray()

            # Reindex dimensions
            _reindexDic = {}
            for t_index in targetIndex:
                _reindexDic.update({t_index.name: t_index.values})
            for coord in dataArray.coords:
                if coord != sourceIndex.name:
                    _reindexDic[coord] = dataArray.coords[coord].values
            _da = _da.reindex(_reindexDic)

            return _da.fillna(0)
        else:
            raise ValueError(
                'mapInfo and targetIndex must have the same number of elements')

    def choice(self, index, selection, includeAll=False):
        """DEPRECATED: Use selector instead. 
            Return the element in the "selection" position of the index. 
        """
        if selection == 0 and includeAll == 1:
            return "All"
        else:
            values = None
            if isinstance(index, pd.Index):
                values = (index.values[:1000]).tolist()
            elif isinstance(index, np.ndarray):
                values = (index[:1000]).tolist()
            else:
                values = list(index)[:1000]
            if not values is None and len(values) >= selection:
                return values[selection-1]
        return ""

    def dynamic(self, dataArray, index, shift, initialValues=None):
        """Perform cyclic calculations betwwen nodes.
            dataArray: dataArray to perform the ciclyc dependency calculation
            index: Index from dataArray to shift 
            shift: number of elemnts to shift. Can be positive or negative
            initialValues: (optional), initial values to apply to first "shift" elemnts
        """
        _da = dataArray.shift({index.name: (shift*-1)})
        if not initialValues is None:
            _da = _da.fillna(initialValues)
        return _da

    def slice_dataarray(self, dataArray, index, position):
        """Filter dataArray by integer position along the specified index.

            dataArray: dataArray to be filtered
            index: pp.index 
            position: int 
            Ex.
                pp.isel(dataArray1, index1, 0)
        """
        if not isinstance(dataArray, xr.DataArray):
            raise ValueError(
                "the 'dataArray' parameter must be of the type xr.DataArray")
        return dataArray.isel({index.name: position}, drop=True)

    def fill_inf(self, dataArray, value=0):
        """Fill np.inf values with default value
            Ex.
                pp,fill_inf(dataArray,0)
        """
        return self.apply_fn(dataArray, lambda x: value if np.isinf(x) else x)

    def fill_all(self, dataArray, value=0):
        """Fill np.inf and np.nan with default value
            Ex.
                pp.fill_all(dataArray,0)
        """
        return self.fill_inf(dataArray.fillna(value), value)

    def add_periods(self, start, periods, freq='M', format='%Y.%m'):
        """Add periods to a date. Can set freq and output format 
           Ex.
                pp.addPeriods('2016.01',6)
                pp.apply( pp.addPeriods, inicio_de_proyectos , duracin_de_proyectos)
        """
        if "." in start:
            start = start.replace('.', '-')
        if periods < 0:
            return pd.period_range(end=start, periods=-periods+1, freq=freq).strftime(format)[0]
        else:
            return pd.period_range(start=start, periods=periods+1, freq=freq).strftime(format)[-1]

    def npv(self, rate, flow, time_index, offset=1):
        """"Returns the Net Present Value (NPV) of a cash flow with equally spaced periods. The flow parameter must contain a series of periodic payments (negative values) and inflows (positive values), indexed by time_index.
            The optional offset parameter especifies the offset of the first value relative to the current time period. By default, offset is set to 1, indicating that the first value is discounted as if it is one step in the future
        """
        _number_of_periods = self.get_pos(time_index) + offset
        _present_values = flow / (1 + rate) ** _number_of_periods
        _npv = _present_values.sum(time_index.name)
        return _npv

    def copy_index(self, dataArray, sortValues=True):
        """Generates a pd.Index with the unique values of the dataArray.
        """
        np_values = dataArray.values.flatten()

        # Numpy unique function automatically reorders. Pandas unique, does not.
        if sortValues is True:
            return pd.Index(np.unique(np_values))
        else:
            return pd.Index(np_values).unique()

    def sequence_index(self, _start, _end, _step=1):
        """
        Returns a pd.Index with the sequence between 'start' and 'end' parameters. Both limits are inclusive. Values are converted to string.
        """
        try:
            _start = int(_start)
            _end = int(_end) + 1
            _step = int(_step)
        except:
            raise ValueError(
                "Only numbers are allowed as 'start', 'end' and 'step' parameters")

        _list = [str(x) for x in range(_start, _end, _step)]
        _index = pd.Index(_list)
        return _index

    def subindex(self, dataArray, targetValue, targetIndex, method='Last'):
        """ Returns a dataArray containing the value of targetIndex for which dataArray (indexed by targetIndex) is equal to targetValue.
            dataArray: Xarray dataArray.
            targetValue: Integer, Float or String.
            targetIndex: Pandas Index.
            method: There are two options: "Last" returns the last occurrence of targetIndex for which dataArray is equal to targetValue. "First" returns the first occurrence.
        """

        # Equals dataArray to targetValue and cumulates it along targetIndex.
        _matriz_1_0 = xr.where(dataArray == targetValue, 1, 0)
        _matriz_1_0_acum = xr.where(
            _matriz_1_0 == 1, _matriz_1_0.cumsum(targetIndex.name), 0)

        if method == 'Last':
            # Get max cumulated value along targetIndex
            _max = _matriz_1_0_acum.max(targetIndex.name)
            _max = xr.where(_max == 0, np.nan, _max)
            _matriz_max = xr.where(
                _matriz_1_0_acum == _max, self.to_dataarray(targetIndex), np.nan)
            return _matriz_max.max(targetIndex.name)
        elif method == 'First':
            # Get min (1) cumulated value along targetIndex
            _matriz_min = xr.where(_matriz_1_0_acum == 1,
                                   self.to_dataarray(targetIndex), np.nan)
            return _matriz_min.max(targetIndex.name)
        else:
            raise ValueError("Insert a valid method")

    def concat_rows(self, array_param, index_param):
        """TODO: add doc"""
        _index = pd.Index([])
        for i in index_param.values:
            _index = self.concat_index(_index, pd.Index(
                array_param.sel({index_param.name: i}, drop=True).values))
        return _index

    def log_task(self, task_state="PROGRESS", task_description=None, task_activity=None, task_info=None):
        """
            Generate log entry. Used for schedule tasks
            params:
            task_state: PROGRESS, INFO, WARNING, FAILURE, RETRY, SUCCESS, REVOKED, STARTED, PENDING, RECEIVED
            task_description: Shot description of task. example: start process
            task_activity: other short description
            task_info: json with more info 
        """
        import json
        _params = {
            "state": task_state,
            "description": task_description,
            "activity": task_activity,
            "info": json.dumps(task_info)}

        res = None

        task_log_endpoint = self.model.getNode("task_log_endpoint").result
        if task_log_endpoint:
            # only used from pyplan_engine
            import requests
            from os import environ
            base_host = environ['PYPLAN_API_HOST'] + task_log_endpoint
            res = requests.post(base_host, data=_params)
        else:
            print(str(_params))

        return res

    def pandas_from_xlsb_file(self, filepath):
        """TODO: add doc"""

        if self._exists_module("pyxlsb"):
            from pyxlsb import open_workbook as open_xlsb
            _df = []
            with open_xlsb(filepath) as wb:
                with wb.get_sheet(1) as sheet:
                    for row in sheet.rows():
                        _df.append([item.v for item in row])

            return pd.DataFrame(_df[1:], columns=_df[0])
        else:
            raise ValueError("pyxlsb library not found")

    def selector(self, options, selected, multiselect=False):
        """ Create UI Pyplan selector for desicion nodes
            Params:
                options: List or pd.Index with available values that can be selected 
                selected: current selected index value's
                multiselect: True to allow multiple selection
        """
        return Selector(options, selected, multiselect)

    def send_message(self, message_text, message_title=None):
        """Send message to UI. Only used with Pyplan UI
            Ex.
                pp.send_message("Process complete!")
        """
        if self.model and self.model.ws:
            self.model.ws.sendMsg(message_text, message_title)

    def progressbar(self, progress, message_text=""):
        """Create and update progress bar. Only used with Pyplan UI
            Ex.
                pp.progressbar(20, "Step 1")
                pp.progressbar(100, "Complete!")
        """
        if self.model and self.model.ws:
            self.model.ws.progressbar(progress, message_text)

    def create_report(self, reportItems, reportIndexName="Report index", reportIndex=None):
        """ Concatenate the reportItems dic dataarrays along the reportIndex dimension
            reportItems: dict or list with datarrays to concat (must have the same structure)
            reportIndexName: Name of the new ReportIndex dimension
            reportIndex: Overwrite ReportIndex dimension
            Ex.
                pp.create_report(reportItems={"Demand":demand, "Product Stock":stock}, reportIndexName="New Report" )
        """
        if isinstance(reportItems, dict):
            report_index = list(reportItems)
            report_values = list(reportItems.values())
            _titles = [str(xx.name) for xx in report_values]
            _index = pd.Index(report_index, name=reportIndexName)
            return xr.concat(report_values, _index)
        else:
            _titles = [str(xx.name) for xx in reportItems]
            _index = None
            if reportIndex is None:
                _index = pd.Index(_titles, name=reportIndexName)
            else:
                _index = reportIndex
            return xr.concat(reportItems, _index)

    def pandas_from_dataarray(self, dataarray):
        """Crate dataframe pandas from datarray with n dimensions
            Ex.
                pp.pandas_from_dataarray(dataArrayNode)
        """
        return dataarray.stack(z=dataarray.dims).to_dataframe("value")

    def pandas_from_access(self):
        """Class to manage access databases"""
        return Pandas_from_acc()


class Selector(object):
    """ Class to manage UI Pyplan selectors.
    """

    SERIALIZABLE_PROPERTIES = ['options', 'selected', 'multiselect']

    def __init__(self, options, selected, multiselect=False):
        """ Create UI Pyplan selector for desicion nodes
        Params:
            options: List or pd.Index with available values that can be selected 
            selected: current selected index value's
            multiselect: True to allow multiple selection
        """
        self._options = options
        self._multiselect = multiselect
        self.selected = selected

    @property
    def value(self):
        if self.multiselect:
            return [self.options[i] for i in self.selected]
        else:
            return self.options[self.selected]

    @property
    def options(self):
        return self._options

    @property
    def multiselect(self):
        return self._multiselect

    @property
    def selected(self):
        res = None
        if self.multiselect:
            res = []
            for nn in self._selected:
                if nn < len(self._options):
                    res.append(nn)
            if len(res) == 0:
                res = list(range(len(self._options)))
        else:
            res = self._selected if self._selected < len(self._options) else 0

        return res

    @selected.setter
    def selected(self, value):
        if self.multiselect:
            if value is None:
                self._selected = []
            elif isinstance(value, list):
                self._selected = value
            else:
                self._selected = [value]
        else:
            if isinstance(value, list):
                self._selected = value[0]
            else:
                self._selected = value

    def toObj(self):
        res = dict()
        for k in Selector.SERIALIZABLE_PROPERTIES:
            if hasattr(self, k):
                if k == "options" and isinstance(getattr(self, k), pd.Index):
                    res[k] = getattr(self, k).tolist()
                else:
                    res[k] = getattr(self, k)
        return res

    def isSameValue(self, value):
        if self.multiselect and isinstance(self.selected, list) and isinstance(value, list):
            l1 = self.selected.copy()
            l2 = value.copy()
            l1.sort()
            l2.sort()
            return l1 == l2
        else:
            return self.selected == value

    def generateDefinition(self, definition, value):

        if self.multiselect:
            if not isinstance(value, list):
                if value is None:
                    value = 0
                value = list(value)
            elif len(value) == 0:
                value = list(range(len(self.options)))
        newPos = str(value)

        reg = r'(?:[^\]\[,]+|\[[^\]\[]+\])'
        groups = re.findall(reg, definition)
        if len(groups) > 2:
            if not str(groups[-1]) in ["False)", "True)", "multiselect=False)", "multiselect=True)"]:
                groups.append("False)")
            newDef = ""
            for nn in range(len(groups)-2):
                newDef += groups[nn]
            newDef = f"{newDef},{newPos},{groups[-1]}"
            return newDef
        return None


class Pandas_from_acc():
    """
    Class that allows to read access files with pandas

    EXAMPLES OF USE:

    # Listing the tables.
    for tbl in pandas_from_access.list_tables("my.mdb"):
        print(tbl)

    # Read a small table.
    df = pandas_from_access.read_table("my.mdb", "MyTable")

    # Read a huge table.
    accumulator = []
    for chunk in pandas_from_access.read_table("my.mdb", "MyTable", chunksize=10000):
        accumulator.append(f(chunk))
    """

    TABLE_RE = re.compile("CREATE TABLE \[([a-zA-Z_0-9 ]+)\]\s+\((.*?\));",
                          re.MULTILINE | re.DOTALL)
    DEF_RE = re.compile("\s*\[(\w+)\]\s*(.*?),")

    @classmethod
    def list_tables(cls, rdb_file, encoding="latin-1"):
        """
        :param rdb_file: The MS Access database file.
        :param encoding: The content encoding of the output. I assume `latin-1`
            because so many of MS files have that encoding. But, MDBTools may
            actually be UTF-8.
        :return: A list of the tables in a given database.
        """
        tables = cls.__get_tables(rdb_file, encoding)
        return [table for table, _ in tables]

    @classmethod
    def read_schema(cls, rdb_file, encoding='utf8'):
        """
        :param rdb_file: The MS Access database file.
        :param encoding: The schema encoding. I'm almost positive that MDBTools
            spits out UTF-8, exclusively.
        :return: a dictionary of table -> column -> access_data_type
        """
        output = subprocess.check_output(['mdb-schema', rdb_file])
        lines = output.decode(encoding).splitlines()
        schema_ddl = "\n".join(l for l in lines if l and not l.startswith('-'))
        tables = cls.__get_tables(rdb_file, encoding)
        schema = {}
        for table, defs in tables:
            schema[table] = cls.__extract_defs(defs)

        return schema

    @classmethod
    def to_pandas_schema(cls, schema, implicit_string=True):
        """
        :param schema: the output of `read_schema`
        :param implicit_string: mark strings and unknown dtypes as `np.str_`.
        :return: a dictionary of table -> column -> np.dtype
        """
        pd_schema = {}
        for tbl, defs in schema.items():
            pd_schema[tbl] = None
            sub_schema = {}
            for column, data_type in defs.items():
                dtype = cls.__extract_dtype(data_type)
                if dtype is not None:
                    sub_schema[column] = dtype
                elif implicit_string:
                    sub_schema[column] = np.str_
            pd_schema[tbl] = sub_schema
        return pd_schema

    @classmethod
    def read_table(cls, rdb_file, table_name, *args, **kwargs):
        """
        Read a MS Access database as a Pandas DataFrame.

        Unless you set `converters_from_schema=False`, this function assumes you
        want to infer the schema from the Access database's schema. This sets the
        `dtype` argument of `read_csv`, which makes things much faster, in most
        cases. If you set the `dtype` keyword argument also, it overrides
        inferences. The `schema_encoding keyword argument passes through to
        `read_schema`. The `implicit_string` argument passes through to
        `to_pandas_schema`.

        I recommend setting `chunksize=k`, where k is some reasonable number of
        rows. This is a simple interface, that doesn't do basic things like
        counting the number of rows ahead of time. You may inadvertently start
        reading a 100TB file into memory. (Although, being a MS product, I assume
        the Access format breaks after 2^32 bytes -- har, har.)

        :param rdb_file: The MS Access database file.
        :param table_name: The name of the table to process.
        :param args: positional arguments passed to `pd.read_csv`
        :param kwargs: keyword arguments passed to `pd.read_csv`
        :return: a pandas `DataFrame` (or, `TextFileReader` if you set
            `chunksize=k`)
        """
        if kwargs.pop('converters_from_schema', True):
            specified_dtypes = kwargs.pop('dtype', {})
            schema_encoding = kwargs.pop('schema_encoding', 'utf8')
            schemas = cls.to_pandas_schema(cls.read_schema(rdb_file, schema_encoding),
                                           kwargs.pop('implicit_string', True))
            dtypes = schemas[table_name]
            dtypes.update(specified_dtypes)
            if dtypes != {}:
                kwargs['dtype'] = dtypes

        cmd = ['mdb-export', rdb_file, table_name]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        return pd.read_csv(proc.stdout, *args, **kwargs)

    # private class methods
    @classmethod
    def __get_tables(rdb_file, encoding='utf8'):
        output = subprocess.check_output(['mdb-schema', rdb_file])
        lines = output.decode(encoding).splitlines()
        schema_ddl = "\n".join(l for l in lines if l and not l.startswith('-'))
        return Pandas_from_acc.TABLE_RE.findall(schema_ddl)

    @classmethod
    def __extract_dtype(data_type):
        # Note, this list is surely incomplete. But, I only had one .mdb file
        # at the time of creation. If you see a new data-type, patch-pull or just
        # open an issue.
        data_type = data_type.lower()
        if data_type.startswith('double'):
            return np.float_
        elif data_type.startswith('long'):
            return np.float_
        elif data_type.startswith('bool'):
            return np.bool_
        elif data_type.startswith('text') or data_type.startswith('memo'):
            return np.str_
        elif data_type.startswith('ole'):
            return np.bytes_
        else:
            return None
    @classmethod
    def __extract_defs(defs_str):
        defs = {}
        lines = defs_str.splitlines()
        for line in lines:
            m = Pandas_from_acc.DEF_RE.match(line)
            if m:
                defs[m.group(1)] = m.group(2)
        return defs



import json
from pyplan_core.classes.evaluators.BaseEvaluator import BaseEvaluator
import plotly.io as pio
from plotly.graph_objs._figure import Figure


class PlotlyEvaluator(BaseEvaluator):

    def evaluateNode(self, result, nodeDic, nodeId, dims=None, rows=None, columns=None, summaryBy="sum", bottomTotal=False, rightTotal=False, fromRow=0, toRow=0):
        fig = result
        if not "layout" in fig or not "margin" in fig["layout"] or fig["layout"]["margin"]["b"] is None:
            fig.update_layout(margin=dict(b=10))
        if not "layout" in fig or not "margin" in fig["layout"] or fig["layout"]["margin"]["t"] is None:
            fig.update_layout(margin=dict(t=10))
        if not "layout" in fig or not "margin" in fig["layout"] or fig["layout"]["margin"]["l"] is None:
            fig.update_layout(margin=dict(l=10))

        res = dict()
        res["result"] = pio.to_html(
            fig, full_html=False, include_plotlyjs=False)
        return json.dumps(res)



from pyplan_core.classes.wizards.BaseWizard import BaseWizard
import pandas as pd
import json


class Wizard(BaseWizard):

    def __init__(self):
        self.code = "IndexFromPandas"

    def generateDefinition(self, model, params):
        nodeId = params["nodeId"]
        res = ""

        if model.existNode(nodeId):
            base_node = model.getNode(nodeId)

            for column_name in params["indexes"]:
                new_index_id = model._removeDiacritics(column_name)
                if model.existNode(new_index_id):
                    new_index_id = model.getNextIdentifier(new_index_id)

                node_position = model.getAPlace(base_node.moduleId, int(
                    base_node.x) + 150, int(base_node.y), 96, 48)
                new_index_node = model.createNode(
                    new_index_id, nodeClass="index", moduleId=base_node.moduleId, x=node_position["x"], y=node_position["y"])

                new_index_node.title = column_name
                new_index_node.definition = f"result = pd.Index({nodeId}.reset_index()['{column_name}'].unique().tolist())"
                res = new_index_node.identifier

        return res
        



import inspect
import numbers
import os
from sys import exc_info, getsizeof

import numpy as np
import numpydoc
import pandas as pd

from pyplan_core import cubepy 
import pyplan_core.cubepy.factory as cpf


class Helpers(object):

    random = -123798
    exact = 1
    start_with = 2
    end_with = 3
    contain = 4
    byVal = 1
    byPos = 2

    @property
    def currentNodeId(self):
        return self.node.identifier

    def __init__(self, node):
        self.node = node

    def release(self):
        self.node = None

    def cube(self, axes, values=None, broadcast=True, dtype=None):
        """Create a cube  object. 
        axes: list of axis of the cube
        values: optional, list of values of the cube. Can be other cubes for build a report.
        Ex.
            cp.cube([time])
            cp.cube([time,product])
            cp.cube([time,product],[10,20,30,40,50,60,70,80])
            cp.cube([time,product],cp.random)

            cp.cube([index_reports],[report_1,report_2])
        """
        return cpf.cube(axes, values, broadcast, dtype)

    def index(self, values):
        """Create a index object.
        values: list of values of the index.
        Ex.
            cp.index(["Item 1","Item 2","Item 3"])
            cp.index([2016,2017,2018])
        """
        return cpf.index(self.node.identifier, values)

    def find(self, param1, param2, compareType=1, caseSensitive=True):
        """
        param1: value or indexarray for compare
        param2: index compare to
        compareType: cp.exact=1, cp.start_with=2, cp.end_with=3, cp.contain=4  
        caseSensitive: able to differentiate between uppercase and lowercase (by default True)

        If param1 is a scalar (numeric or str) and param2 is an index:  return cube indexed by param2 with True on ocurrences of param2
            Ex. result = cp.apply("te", region, cp.end_with)
        If param1 is an index and param2 is an index too:  return cube indexed by param1 and param2 with True on ocurrences of param1 on param2
            Ex. result = cp.apply(subregion, region, cp.contain)

        """
        return cpf.find(param1, param2, compareType, caseSensitive)

    def selectText(self, data, first=None, last=None):
        """Returns a new cube with the text contained between the 'first' and 'last' characters of cube / index 'data'. Starts counting from 0.
                If 'first' character is ommited, it returns every character from the first character of 'data' to the 'last' character, inclusive.
                If 'last' character is ommited, it returns every character from "first" character of 'data', to the last character available for each cell.
        """
        return cpf.selectText(data, first, last)

    def apply(self, fn, param1, param2=None, start=None):
        """
        Apply functions to index and cubes. Multiple results can be obtained
        fn: function to apply
        param1: index or cube
        param2: index or cube
        start: scalar or cube
            Ex. 
                cp.apply(lambda x: x[:2] ,indexRegion): return new cube indexed by "indexRegion" and apply fn on each item
                cp.apply(lambda x: x*5 ,cubeSales): return new cube result of apply fn on all values of the cubeSales
                cp.apply( cp.addPeriods, start_proj , end_proj): Return new cube result of apply fn on "start_proj" with "end_proj"
                cp.apply(lambda x: x+1, indexYear, start=10) : Return new cube indexed by "indexYear", result of apply fn starting with scalar value "10"
                cp.apply(lambda x: x+1, indexYear, start=prices) : Return new cube indexed by "indexYear", result of apply fn starting with cube "prices"
        """
        return cpf.apply(fn, param1, param2, start)

    def max(self, cube1, cube2):
        """Return max value between two cubes
        """
        return cpf.max(cube1, cube2)

    def min(self, cube1, cube2):
        """Return min value between two cubes
        """
        return cpf.min(cube1, cube2)

    def sum(self, cube, axis=None, keep=None, group=None, sort_grp=True):
        """Sum of array elements over a given axis.
        :param axis: Axis or axes along which a sum is performed. The default (axis = None) is perform a sum
        over all the dimensions of the input array. axis may be negative, in which case it counts from the last
        to the first axis. If this is a tuple of ints, a sum is performed on multiple axes, instead of a single
        axis or all the axes as before.
        :return: new Cube instance or a scalar value
        """
        return cpf.sum(cube, axis, keep, group, sort_grp)

    def subscript(self, cube, index, value):
        """Filter cube1 using the index and the value. Return a new cube without the index1 dimension
            Ex.
                cp.subscript(nodo_ejemplo,index_para_ejemplo,"Item 1")
        """
        if isinstance(index, str):
            index = self.node.model.getNode(index)
        return cpf.subscript(cube, index, value)

    def slice(self, cube, index, value):
        """Filter cube1 using the index and the value. Return a new cube without the index1 dimension
            Ex.
                cp.slice(nodo_ejemplo,index_para_ejemplo,2)
        """
        return cpf.slice(cube, index, value)

    def shift(self, cube, axis, n=1, cval=0):
        """Returns a cube with the axis shifted.
            Ex.
                cp.shift(nodo_ejemplo,index_para_ejemplo,1)
        """
        return cpf.shift(cube, axis, n, cval)

    def subset(self, cube):
        """Returns a list of all the elements of the index for which cube is true. The function is used to create a new index that is a subset of an existing index.
            Ex. cp.subset(cantidades>0)
        """
        return cpf.subset(cube, self.currentNodeId)

    def aggregate(self, cube, mapCube, indexToRemove, targetIndex):
        """ Aggregates the values in Cube to generate the result indexed by  targetIndex.
            Map gives the value of targetIndex for each element of indexToRemove

            Example for aggregating time information into annual index the syntax is:
                cp.aggregate(cube, map, time, years )
        """
        return cpf.aggregate(cube, mapCube, indexToRemove, targetIndex)

    def cumulate(self, cube, index):
        """ TODO coment
        """
        return cpf.cumulate(cube, index)

    def cumProd(self, cube, index):
        """Return the cumulative product of elements along a given axis
            param cube: cube 
            param axis: axis name (str), index (int) or instance
            Ex:
                cp.cumProd(nodo,indice)
        """
        return cpf.cumProd(cube, index)

    def irr(self, flow, time_index):
        """Returns the Internal Rate of Return (IRR) of a series of periodic payments (negative values) and inflows (positive values). The IRR is the discount rate at which the Net Present Value (NPV) of the flows equals zero. 
            The variable flow must be indexed by time_index.

        If the cash flow never changes sign, cp.irr() has no solution and returns NAN (Not A Number).
        """

        return cpf.irr(flow, time_index)

    def npv(self, rate, flow, time_index, offset=1):
        """"Returns the Net Present Value (NPV) of a cash flow with equally spaced periods. The flow parameter must contain a series of periodic payments (negative values) and inflows (positive values), indexed by time_index.
            The optional offset parameter especifies the offset of the first value relative to the current time period. By default, offset is set to 1, indicating that the first value is discounted as if it is one step in the future
        """
        return cpf.npv(rate, flow, time_index, offset)

    # Esta es una nueva forma de hacer aggregate, pero falla en algunas ocasiones. ver con Lucas
    # def aggregate(self, cube, mapCube, indexToRemove, targetIndex,  ):

    #     _model = self.node.model
    #     _mapCube_list = []

    #     if type( mapCube ) is list:
    #         for _cube in mapCube:
    #             _mapCube_list.append( _cube )
    #     else:
    #         _mapCube_list = [ mapCube ]

    #     # Transforms set of cubes into a list.
    #     _indexToRemove_st = indexToRemove.name

    #     # Transforms set of indexes into a list.
    #     _targetIndex_st = []
    #     if type( targetIndex ) is list:
    #         for _ind in targetIndex:
    #             _targetIndex_st.append( _ind.name )
    #     else:
    #         _targetIndex_st = [ targetIndex.name ]

    #     # Creates pairs of ( mapCube, targetIndex ).
    #     _map_index_pairs = []
    #     for _pos in range( 0, len( _mapCube_list ) ):
    #         _pair = [ _mapCube_list[_pos], _targetIndex_st[_pos] ]
    #         _map_index_pairs.append( _pair )

    #     # Creates a dataframe with a indexToRemove-to-targetIndex allocation.
    #     _df_allocation = pd.DataFrame()
    #     for nn, pair in enumerate( _map_index_pairs ):
    #         #_allocation = self.pandasFromCube( _model.getNode(pair[0]).result , dataColumnName = pair[1], keepIndexOrder = False ).reset_index()
    #         _allocation = self.pandasFromCube( pair[0] , dataColumnName = pair[1], keepIndexOrder = False ).reset_index()

    #         if nn == 0:
    #             _df_allocation = _allocation
    #         else:
    #             _df_allocation = _df_allocation.merge( _allocation , on = _indexToRemove_st , how = "outer" )

    #     _df = self.pandasFromCube( cube , dataColumnName = 'value', keepIndexOrder = False ).reset_index()
    #     _df = _df[_df['value'] != 0]
    #     _df = _df.merge( _df_allocation , on = _indexToRemove_st , how = 'left')

    #     # Eliminates columns that will not be part of the cube's dimensions.
    #     _columns = list( _df.columns.values )
    #     _columns.remove( 'value' )
    #     _columns.remove( _indexToRemove_st )

    #     _df = _df.groupby( _columns ).agg({"value": sum})

    #     #remove duplicates
    #     _finalColumns = []
    #     for _tmpIndex  in _columns:
    #         if (_tmpIndex.endswith("_x") or _tmpIndex.endswith("_y")):
    #             _finalColumns.append(_tmpIndex[:-2])
    #         else:
    #             _finalColumns.append(_tmpIndex)
    #     _finalColumns = set(_finalColumns)
    #     _finalColumns = list(_finalColumns)

    #     _index = [ _model.getNode(idx).result for idx in _finalColumns]
    #     _cube = self.cubeFromPandas( _df, _index, "value" )

    #     return _cube

    def isNumber(self, value):
        """Return true if value is number
        """
        try:
            float(value)
            return True
        except ValueError:
            return False

    def splitText(self, text, separator=",", resultIndex=None):
        """Split text into a list of substrings by each occurrence of separator.
            text:  string or cp.index used for split
            separator: string for use as separator
            resultIndex: optional, for return cp.cube indexes by this index
            Ex:
                cp.splitText("Uno Dos Tres Cuatro Cinco Seis", separator=" ") 
                cp.splitText(indice_para_split, separator="-") 
                cp.splitText("Uno Dos Tres Cuatro Cinco Seis"," ",result_index)  
                cp.splitText(indice_para_split, "-", result_index)  

        """
        return cpf.splitText(text, separator, resultIndex, self.currentNodeId)

    def changeIndex(self, cube, oldIndex, newIndex, compareMode=1, default=None):
        """Returns a new Cube instance with the axes changed,looking newIndex on oldIndex by value o by position.
            param Comparemode: cp.byVal or cp.byPos
            Ex:
                cp.changeIndex(nodo_original,indice_original,indice_cambiado,cp.byVal)
                cp.changeIndex(nodo_original,indice_original,indice_cambiado,cp.byPos)
        """
        return cpf.changeIndex(cube, oldIndex, newIndex, compareMode, default)

    def copyIndex(self, cube):
        """Generate a cp.index with current unique values of the cube. 
        The cube must have only one dimension
        """
        return cpf.copyIndex(cube, self.currentNodeId)

    def cascadeVolume(self, demand, ranges, consumption_range_index=None):
        """TODO Comment
        Ex:
                cp.cascadeVolume(cantidades,limites_rango_consum)
                cp.cascadeVolume(cantidades,limites_rango_consum,"rangos_consumo")
                cp.cascadeVolume(cantidades,limites_rango_consum,rangos_consumo)
        """
        return cpf.cascadeVolume(demand, ranges, consumption_range_index)

    def bandAllocation(self, demand, ranges, consumption_range_index=None):
        """TODO Comment
            Ex:
                cp.bandAllocation(cantidades,limites_rango_consum)
                cp.bandAllocation(cantidades,limites_rango_consum,"rangos_consumo")
                cp.bandAllocation(cantidades,limites_rango_consum,rangos_consumo)
        """
        return cpf.bandAllocation(demand, ranges, consumption_range_index)

    def dispatch(self, contract_vol, contract_price, contract_index, demand):
        """TODO comment
            Ex:
                cp.dispatch(contract_volumes,prices,contracts,demand)
        """
        return cpf.dispatch(contract_vol, contract_price, contract_index, demand)

    def argsort(self, cube, axis):
        """Return new cube with the position of values sorted by axis
            ex.
                cp.argsort(nodo_1,index_a)
        """
        return cpf.argsort(cube, axis)

    def iif(self, condition, truePart, falsePart=None):
        """Inline if. Evaluate condition and return truePart or fasePart
            ex. 
                cp.iif( producto=="Producto A", cantidades*1000 , cantidades)
                cp.iif( cantidades > 10, cantidades , 0)
        """
        return cpf.iif(condition, truePart, falsePart)

    def mdtable(self, dataframe, cubeIndexes, valueColumns, indexColumnHeaders=None, replaceByIndex=None):
        """
        DEPRECATED: Use "cp.cubeFromPandas" instead
        """
        return self.cubeFromPandas(dataframe, cubeIndexes, valueColumns, indexColumnHeaders, replaceByIndex)

    def cubeFromPandas(self, dataframe, cubeIndexes, valueColumns, indexColumnHeaders=None, replaceByIndex=None):
        """Creat new cp.cube, converting pandas to multidimensional data, according to the parameters
            dataframe: pandas dataframe
            cubeIndexes: objects cp.index for perform change index
            valueColumns: string with column name of the dataframe where contain the values
                            cp.index with columns names for convert colums to index
            indexColumnHeaders: (optional) column names in pandas to parse with cubeIndexes. Used if header on dataframe is not equal to index identifiers.
            replaceByIndex: (optional) replace index used in valueColumns for this index. (using changeindex)
        """

        return cpf.cubeFromPandas(dataframe, cubeIndexes, valueColumns, indexColumnHeaders, replaceByIndex)

    def indexFromPandas(self, dataframe, columnName=None, removeEmpty=True):
        """ Return a cp.index from an columnf of a pandas dataframe.
        dataframe: pandas dataframe
        columnName: dataframe column name used for create cp.index. By default is created using the first column
        removeEmpty: True for remove empty rows
            Ex.
                cp.indexFromPandas(df)
                cp.indexFromPandas(df,"column10")
        """
        return cpf.indexFromPandas(dataframe, columnName, removeEmpty, self.currentNodeId)

    def pandasFromCube(self, cube, dataColumnName="value", keepIndexOrder=False):
        """
        Return indexed Dataframe created from cubePy Cube
        """
        return cpf.pandasFromCube(cube, dataColumnName, keepIndexOrder)

    def excel(self, filepath, useOpenpyxl=False, dataOnly=True, readOnly=True):
        """ Create excel object from filepath.
        filepath: absolute or relative path to excel file
        useOpenpyxl: True for use custom 
        dataOnly: True for view only the values, not formula
        readOnly: True for read only, False for write options
        Ex.
                cp.excel("\path\to\the\excelfile.xlsx")
        """
        fullFilename = filepath

        if not os.path.isfile(fullFilename):
            fullFilename = self.node.model.getNode(
                "current_path").result + filepath
        return cpf.excel(fullFilename, useOpenpyxl, dataOnly, readOnly)

    def pandasFromExcel(self, excel, sheetName=None, namedRange=None, cellRange=None, indexes=None, driver='Driver={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)};DBQ=%s;READONLY=TRUE'):
        """ return a pandas dataframe from excel.
        excel: path to excel file or cp.excel object
        sheetName: sheet name to be read
        namedRange: name of the range to be read
        cellRange: used with sheetname, for read from a simple range
        indexes: Listo of columns names for convert to index of dataframe
            Ex.
                cp.pandasFromExcel(excelNode,"Sheet 1")
                cp.pandasFromExcel(excelNode,namedRange="name_range")
                cp.pandasFromExcel(excelNode,"Sheet 1",cellRange="A1:H10")

        """
        if isinstance(excel, str):

            if not os.path.isfile(excel):
                excel = self.node.model.getNode("current_path").result + excel
        return cpf.pandasFromExcel(excel, sheetName, namedRange, cellRange, indexes, driver)

    def indexFromExcel(self, excel, sheetName=None, namedRange=None, cellRange=None, columnName=None, removeEmpty=True):
        """ Return a cp.index from an excel file.
        excel: cp.excel object
        sheetName: sheet name to be read
        namedRange: name of the range to be read
        cellRange: used with sheetname, for read from a simple range
        columnName: dataframe column name used for create cp.index. By default is created using the first column
        removeEmpty: True for remove empty rows
            Ex.
                cp.indexFromExcel(excelNode,"Sheet 1")
                cp.indexFromExcel(excelNode,namedRange="name_range")
                cp.indexFromExcel(excelNode,namedRange="name_range", columnName="indicadores")
        """
        return cpf.indexFromExcel(excel, sheetName, namedRange, cellRange, columnName, removeEmpty, self.currentNodeId)

    def cubeFromExcel(self, excel, sheetName=None, namedRange=None, cellRange=None, cubeIndexes=None, valueColumns=None, indexColumnHeaders=None, replaceByIndex=None):
        """ Return a cp.cube from excel file.
         excel: cp.excel object
        sheetName: sheet name to be read
        namedRange: name of the range to be read
        cellRange: used with sheetname, for read from a simple range
        cubeIndexes: objects cp.index for perform change index
        valueColumns: string with column name of the dataframe where contain the values
                        cp.index with columns names for convert colums to index
        indexColumnHeaders: (optional) column names in pandas to parse with cubeIndexes. Used if header on dataframe is not equal to index identifiers.
        replaceByIndex: (optional) replace index used in valueColumns for this index. (using changeindex)

            Ex.
                cp.cubeFromExcel(excelNode,"Sheet 1",cubeIndexes=[indicadores],valueColumns="descuentos")
                cp.cubeFromExcel(excelNode,namedRange="nombre_rango",cubeIndexes=[indicadores],valueColumns=time)
        """
        return cpf.cubeFromExcel(excel, sheetName, namedRange, cellRange, cubeIndexes, valueColumns, indexColumnHeaders, replaceByIndex)

    def sequence(self, initialNum, finalNum, stepSize=1, dtype=None):
        """
        Creates an index with a list of numbers increasing or decreasing from initial_num to final_num by increments (or decrements) of stepSize, which is optional and defaults to 1
        """
        return cpf.sequence(initialNum, finalNum, stepSize, dtype, self.currentNodeId)

    def lookup(self, cubeWithData, cubeWithMap, sharedIndex):
        """
        Returns the value of cubeWithData indexed by the index of cubeWithMap.
        cubeWithData must be indexed by sharedIndex and cubeWithData values must correspond to elements of sharedIndex.

        For example: Let's say you have a cube with an estimated inflation rate by Country ("inflation_rate" is the name of the cube; "country" is the name of the index) and you want to assign it to the corresponding Company depending on its location. On the other hand, there's a many-to-one map where each Company is allocated to a single Country ("country_to_company_allocation"). The sharedIndex, in this case, is Country ("country").
        As a result, 
            cp.lookup( inflation_rate , country_to_company_allocation , country )
        will return the estimated inflation rate by Company.
        """
        return cpf.lookup(cubeWithData, cubeWithMap, sharedIndex)

    def cubeFromNumpy(self, npArray):
        """ 
        Return a cube object from numpy Array. Generate temporal indexes
        """
        return cpf.cubeFromNumpy(npArray)

    def movingStats(self, cube, index, window=12, fn=np.mean):
        """
        Allows to calculate in a moving Window the selected statistics measure
        """
        return cpf.movingStats(cube, index, window, fn)

    # Cubepy Methods

    def plot(self, fig, **kwargs):
        """
        Generate HTML with plotly chart
        Params:
            fig: plotly fig
        """
        import plotly.io as pio
        return pio.to_html(fig, full_html=False, **kwargs)

    def concatIndex(self, *args):
        """
        Concatenates two or more indexes and/or atomic values into a single new index
        Return: new index
            Ex.
                cp.concatIndex(index1,index2,index3,value1,value2)
        """
        _list = []
        for arg in args:
            if isinstance(arg, cubepy.Index):
                values = (arg.values).tolist()
                _list.extend(values)
            else:
                _list.append(arg)

        seripandas = pd.Series(_list)
        return self.index(seripandas.unique())

    def yearsToTime(self, cube, yearsIndex, timeIndex, div=12):
        """ Convert cube indexed by anual index to a monthly index
        cube are the cp.cube with annual values to be converted
        yearIndex: is the year index of the original matrix
        timeIndex: is the monthly index of the result
        div: is an optional parameter used as divisor of the annual values (the tipical value is 12)
        """
        matrix = self.find(yearsIndex, timeIndex, compareType=4)
        final_cube = self.sum(cube * matrix, axis=[yearsIndex]) / div
        return final_cube

    def dynamic(self, cube, index, shift, initialValues=None):
        """
        Perform cyclic calculations betwwen nodes.
        cube: cp.cube to evaluate
        index: Index to shift 
        shift: amount of elemnts to shift. Can be positive or negative
        initialValues: (optional), initial values to apply to first "shift" elemnts
        """
        return self.shift(cube, index, shift, initialValues)

    def addPeriods(self, start, periods, freq='M', format='%Y.%m'):
        """Add periods to a date. Can set freq and output format 
            Ex.
                cp.addPeriods('2016.01',6)
                cp.apply( cp.addPeriods, inicio_de_proyectos , duracin_de_proyectos)
        """
        if "." in start:
            start = start.replace('.', '-')
        if periods < 0:
            return pd.period_range(end=start, periods=-periods+1, freq=freq).strftime(format)[0]
        else:
            return pd.period_range(start=start, periods=periods+1, freq=freq).strftime(format)[-1]

    def createTime(self, date_start, date_end, freq='M', format='%Y.%m'):
        """Create time index usign start and end dates and freq. The result is formated to format parameter
            Ex.
                cp.createTime('2016.01','2018.12')
                cp.createTime('2016.01','2016.12',freq='D',format='%d/%m/%Y')
        """
        if "." in date_start:
            date_start = date_start.replace('.', '-')
        if "." in date_end:
            date_end = date_end.replace('.', '-')
        return self.index(pd.period_range(start=date_start, end=date_end, freq=freq).strftime(format))

    def choice(self, index, selection, includeAll=False):
        """Return the element in the "selection" position of the index. 
        """
        if selection == 0 and includeAll == 1:
            return "All"
        else:
            values = None
            if isinstance(index, cubepy.Index):
                values = (index.values[:1000]).tolist()
            elif isinstance(index, np.ndarray):
                values = (index[:1000]).tolist()
            else:
                values = list(index)[:1000]
            if not values is None and len(values) >= selection:
                return values[selection-1]
        return ""

    def copyAsValues(self, source, targetId):
        """Copy values of cube "source" into cube with id 'targetId'. This function alter the thefinition of cube with 'targetId' identifier.
            source: cube/index from copy values
            targetId: identifier (string) of the target node 
        """

        if isinstance(source, str):
            if self.node.model.existNode(source):
                source = self.node.model.getNode(source)
            else:
                raise ValueError("Node '" + source + "' not found")

        if not isinstance(source, cubepy.Cube) and not isinstance(source, cubepy.Index):
            raise ValueError(
                "The 'source' parameter must be a cp.cube or cp.index object")

        if not isinstance(targetId, str):
            raise ValueError(
                "The 'targetId' parameter must be a string (identifier of node)")

        if not self.node.model.existNode(targetId):
            raise ValueError("Node '" + targetId + "' not found")

        self.node.model.getNode(
            targetId).definition = source.generateDefinition()
        return True

    def goalSeek(self, nodeIdX, nodeIdObjective, goal=0, startValue=1):
        """ Finds the value of nodeIdX that makes nodeIdObjective equal to goal.
        nodeIdX: String with id of node X
        nodeIdObjective: String with id of node X
        """
        from scipy.optimize import newton

        def _f(x):
            self.node.model.getNode(nodeIdX).definition = "result = " + str(x)
            value = self.node.model.getNode(nodeIdObjective).result
            return value - goal

        _res = newton(_f, x0=startValue)
        return _res

    def from_xindex(self, xindex):
        """Create a cubepy index from xarray index 
        """
        return cpf.index(self.node.identifier, xindex.values)

    def from_dataArray(self, axes, dataArray):
        """Create a cubepy cube from dataarray 
        """
        return self.cube(axes, values=dataArray.values)

    @staticmethod
    def getResultSize(result):
        res = getsizeof(0)
        res += getsizeof(result)
        return res

    @staticmethod
    def kindToString(kind):
        return cubepy.cube.kindToString(kind)
        """
        if kind in {'U', 'S'}:
            return "string"
        elif kind in {'b'}:
            return "boolean"
        elif kind in {'i','u','f','c'}:
            return "numeric"
        elif kind in {'m','M'}:
            return "date"
        elif kind in {'O'}:
            return "object"
        elif kind in {'V'}:
            return "void"
        """

    @staticmethod
    def describe(text, prefix, objectType=None):
        res = []

        _prefix = prefix + "."
        if text == _prefix:
            text = ""
            _prefix = ""

        _members = inspect.getmembers(objectType)
        for _member in _members:
            if not("__" in str(_member[0])) and (text is None or text in str(_member[0]).lower()):
                if inspect.isfunction(_member[1]):
                    _doc = inspect.getdoc(_member[1])
                    if not _doc is None:
                        _params = str(inspect.signature(_member[1]))
                        _params = _params.replace("self,", "")
                        res.append({
                            "identifier": _prefix + _member[0],
                            "nodeClass": "helper" if prefix == "cp" else "function",
                            "moduleId": "",
                            "title": _prefix + _member[0],
                            "description": _doc,
                            "params": _params
                        })
                elif inspect.isroutine(_member[1]):
                    _doc = inspect.getdoc(_member[1])
                    try:
                        _params = str(inspect.signature(_member[1]))
                        _params = _params.replace("self,", "")
                        res.append({
                            "identifier": _prefix + _member[0],
                            "nodeClass": "method",
                            "moduleId": "",
                            "title": _prefix + _member[0],
                            "description": _doc,
                            "params": _params
                        })
                    except:
                        # check if object is form numpy class
                        _dot = str(objectType).find('.')
                        if _dot != -1 and str(objectType)[_dot-5:_dot] == 'numpy':
                            info = numpydoc.docscrape.FunctionDoc(_member[1])
                            info["Signature"] = info["Signature"][info["Signature"].find(
                                "("):]
                            res.append({
                                "identifier": _prefix + _member[0],
                                "nodeClass": "method",
                                "moduleId": "",
                                "title": _prefix + _member[0],
                                "description": _doc,
                                "params":  info["Signature"]
                            })
                        else:
                            print(
                                "Error al intentar obtener la signature del tipo de dato")
                            e = exc_info()[0]
                            print("<p>Error: %s</p>" % e)
        return res

    @staticmethod
    def dfs_topsort(graph):
        """
        DFS algorithm for sort nodes
        """
        L = []
        color = {u: "white" for u in graph}
        found_cycle = [False]
        for u in graph:
            if color[u] == "white":
                Helpers.dfs_visit(graph, u, color, L, found_cycle)
            if found_cycle[0]:
                break

        if found_cycle[0]:           # if there is a cycle,
            L = []                   # then return an empty list

        L.reverse()                  # reverse the list
        return L                     # L contains the topological sort

    @staticmethod
    def dfs_visit(graph, u, color, L, found_cycle):
        """
        DFS algorithm for sort nodes
        """
        if found_cycle[0]:
            return
        color[u] = "gray"
        for v in graph[u]:
            if color[v] == "gray":
                found_cycle[0] = True
                return
            if color[v] == "white":
                Helpers.dfs_visit(graph, v, color, L, found_cycle)
        color[u] = "black"      # when we're done with u,
        L.append(u)             # add u to list (reverse it later!)



import json
import numpy as np
import pandas as pd
import inspect
import jsonpickle
from pyplan_core import cubepy 
from pyplan_core.classes.common.filterChoices import filterChoices
from pyplan_core.classes.common.indexValuesReq import IndexValuesReq


class BaseEvaluator(object):
    """
    Base Class to manage node evaluators
    """

    def createResult(self, result, structure, resultIsJson=False, onRow=None, onColumn=None, node=None, pageInfo=None):
        if resultIsJson:

            res = {
                "resultType": json.dumps(structure),
                "result": json.loads(result),
            }
            if not onRow is None:
                res["onRow"] = onRow
            if not onColumn is None:
                res["onColumn"] = onColumn
            if not pageInfo is None:
                res["pageInfo"] = pageInfo
            if not node is None and node.numberFormat:
                res["numberFormat"] = node.numberFormat

            return json.dumps(res)
        else:

            res = {
                "resultType": json.dumps(structure),
                "result": result,
                "onRow": onRow,
                "onColumn": onColumn
            }
            if not pageInfo is None:
                res["pageInfo"] = pageInfo

            # check if in scenario
            if not node is None:
                if node.model.isNodeInScenario(node.identifier):
                    res["scenario"] = True
                if node.numberFormat:
                    res["numberFormat"] = node.numberFormat

            return json.dumps(res)

    def getStructure(self, result):
        structure = dict()
        structure["type"] = str(type(result))
        return structure

    def checkStructure(self, result, resultType):
        """ Check current vs result structure. Result False for distinct structure """
        res = True
        if resultType:
            try:
                structure = json.loads(resultType)
                result_structure = self.getStructure(result)
                if structure["type"] != result_structure["type"]:
                    res = False
            except Exception as ex:
                print(f"Error checking structure: {ex}")

        return res

    def evaluateNode(self, result, nodeDic, nodeId, dims=None, rows=None, columns=None, summaryBy="sum", bottomTotal=False, rightTotal=False, fromRow=0, toRow=0):
        result_structure = self.getStructure(result)
        if isinstance(result, np.ndarray):
            return self.createResult(result.tolist(), result_structure, node=nodeDic[nodeId])
        elif callable(result):  # is function
            aux = {
                "params": inspect.getargspec(result)[0]
            }
            return self.createResult(jsonpickle.encode(aux), result_structure, node=nodeDic[nodeId])
        else:

            try:
                return self.createResult(result, result_structure, node=nodeDic[nodeId])
            except:
                return self.createResult(str(result), result_structure, node=nodeDic[nodeId])

    def getCubeValues(self, result, nodeDic, nodeId,
                      query): raise NotImplementedError

    def getCubeDimensionValues(
        self, result, nodeDic, nodeId, query): raise NotImplementedError

    def getCubeMetadata(self, result, nodeDic,
                        nodeId): raise NotImplementedError

    def generateEmptyPreviewResponse(self, nodeDic, nodeId):
        msg = "This node does not return any value"
        if nodeDic[nodeId].nodeClass == "button":
            msg = "Successfully executed"
        res = {
            "resultType": "",
            "dims": [],
            "console": nodeDic[nodeId].lastEvaluationConsole,
            "preview": msg
        }
        return json.dumps(res)

    def previewNode(self, nodeDic, nodeId):

        if not nodeDic[nodeId].result is None:
            res = {
                "resultType": str(type(nodeDic[nodeId].result)),
                "dims": [],
                "console": nodeDic[nodeId].lastEvaluationConsole,
                "preview": str(nodeDic[nodeId].result)[:1000]
            }

            if callable(nodeDic[nodeId].result):  # is function
                try:
                    res["preview"] += "\nparams: " + \
                        str(inspect.getargspec(nodeDic[nodeId].result)[0])
                except Exception as ex:
                    # in some cases, inspect throws an error when trying to obtain the parameters of a function
                    pass
            return json.dumps(res)
        else:
            return self.generateEmptyPreviewResponse(nodeDic, nodeId)

    def setNodeValueChanges(self, nodeDic, nodeId, changes):
        value = None
        if not changes is None and len(changes["changes"]) == 1 and len(changes["changes"][0]["filterList"]) == 0:
            value = changes["changes"][0]["definition"]
            value = nodeDic[nodeId].sanitizeDefinition(value)

        if not value is None:
            nodeDic[nodeId].definition = value

    def isTable(self, node):
        return "0"

    def isIndexed(self, result):
        return False

    def getIndexes(self, node, result=None):
        return []

    def getIndexesWithLevels(self, node, result=None):
        return []

    def getIndexValues(self, nodeDic, data: IndexValuesReq, result=None):
        res = []
        if data.node_id:
            if (not data.node_id is None) & (data.node_id in nodeDic):
                node = nodeDic[data.node_id]
                if result is None:
                    result = node.result

                if (f"{data.index_id}.{data.node_id}") in self.getIndexes(node):
                    if self.isIndexed(result):
                        for index in result.index.levels:
                            if index.name == data.index_id:
                                res = index.values.tolist()
                                break
                    else:
                        res = result[data.index_id].unique().tolist()
        else:
            if (not data.index_id is None) & (data.index_id in nodeDic):
                node = nodeDic[data.index_id]
                if result is None:
                    result = node.result
                if isinstance(result, cubepy.Index):
                    res = result.values.tolist()
                elif isinstance(result, np.ndarray):
                    res = result.tolist()
                else:
                    res = list(result)
        if data.text1:
            text1 = data.text1.lower()
            if data.filter == filterChoices.CONTAINS.value:
                res = list(
                    filter(lambda item: text1 in str(item).lower(), res))
            elif data.filter == filterChoices.NOT_CONTAINS.value:
                res = list(
                    filter(lambda item: not text1 in str(item).lower(), res))
        return res

    def getIndexType(self, nodeDic, nodeId, indexId):
        return "S"

    def exportFlatNode(self, nodeDic, nodeId, numberFormat, columnFormat, fileName):
        if columnFormat == "tab":
            columnFormat = "\t"

        decimalSep = "."
        if numberFormat == 'TSPDSC':
            decimalSep = ","

        query = {
            "columns": self.getIndexes(nodeDic[nodeId]),
            "filters": None
        }

        data = self.getCubeValues(
            nodeDic[nodeId].result, nodeDic, nodeId, query)
        realIndexes = [x["field"] for x in data["dims"]]
        realIndexes = realIndexes + ["data_value"]

        dimValues = [x["values"] for x in data["dims"]]

        import itertools
        allCombinations = itertools.product(*dimValues)

        with open(fileName, 'w') as f:
            # headers
            f.write(columnFormat.join(realIndexes) + "\n")
            # data
            nn = 0
            for item in allCombinations:
                aa = 0
                f.write(columnFormat.join(str(e) for e in item) + columnFormat)
                f.write(str(data["values"][nn]).replace(
                    ".", decimalSep) + "\n")
                nn += 1

            f.close()

        return True

    def dumpNodeToFile(self, nodeDic, nodeId, fileName):
        definition = nodeDic[nodeId].definition
        with open(fileName, 'w') as f:
            f.write(definition)
            f.close()

    def geoUnclusterData(self, result, nodeDic, nodeId, rowIndex, attIndex, latField="latitude", lngField="longitude", geoField="geoField",
                         labelField="labelField", sizeField="sizeField", colorField="colorField", iconField="iconField"): raise NotImplementedError

    def postCalculate(self, node, result):
        """Method executed after calculate node
        """
        pass

    def copyAsValues(self, result, nodeDic, nodeId):
        """ Copy node as values """
        raise NotImplementedError

    def checkDateFormat(self, np_array):
        if "date" in np_array.dtype.name:
            def _cut(x): return x[:19]
            vfunc = np.vectorize(_cut)
            return vfunc(np_array.astype(str))
        else:
            return np_array
